

def test_standard_case():
    import copy
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs2 import EXAMPLE_INPUT_DICT, EXAMPLE_CONFIG_DICT
    from sfeprapy.mcs2.mcs2_calc import teq_main, teq_main_wrapper, mcs_out_post
    from sfeprapy.func.mcs_gen import main as gen
    from scipy.interpolate import interp1d
    import numpy as np
    # increase the number of simulations so it gives sensible results
    mcs_input = EXAMPLE_INPUT_DICT
    for k in list(mcs_input.keys()):
        mcs_input[k]['n_simulations'] = 2500
    # increase the number of threads so it runs faster
    mcs_config = copy.copy(EXAMPLE_CONFIG_DICT)
    mcs_config['n_threads'] = 4
    mcs = MCS()
    mcs.define_problem(data=mcs_input, config=mcs_config)
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(teq_main, teq_main_wrapper, mcs_out_post)
    mcs.run_mcs()
    mcs_out = mcs.mcs_out
    teq = mcs_out['solver_time_equivalence_solved'] / 60.
    hist, edges = np.histogram(teq, bins=np.arange(0, 181, 0.5))
    x, y = (edges[:-1] + edges[1:]) / 2, np.cumsum(hist / np.sum(hist))
    teq_at_80_percentile = interp1d(y, x)(0.8)
    print(teq_at_80_percentile)
    target, target_tol = 65, 1
    assert target - target_tol < teq_at_80_percentile < target + target_tol


# test gui version
def test_gui():
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs2.mcs2_calc import teq_main as calc
    from sfeprapy.mcs2.mcs2_calc import teq_main_wrapper as calc_mp
    from sfeprapy.func.mcs_gen import main as gen
    mcs = MCS()
    mcs.define_problem()
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(calc, calc_mp)
    mcs.run_mcs()


# test non-gui version
def test_arg_dict():
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs2 import EXAMPLE_INPUT_DICT, EXAMPLE_CONFIG_DICT
    from sfeprapy.mcs2.mcs2_calc import teq_main as calc
    from sfeprapy.mcs2.mcs2_calc import teq_main_wrapper as calc_mp
    from sfeprapy.mcs2.mcs2_calc import mcs_out_post
    from sfeprapy.func.mcs_gen import main as gen
    for k in list(EXAMPLE_INPUT_DICT.keys()):
        EXAMPLE_INPUT_DICT[k]['n_simulations'] = 5
    EXAMPLE_CONFIG_DICT['n_threads'] = 2
    mcs = MCS()
    mcs.define_problem(data=EXAMPLE_INPUT_DICT, config=EXAMPLE_CONFIG_DICT)
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(calc, calc_mp, mcs_out_post)
    mcs.run_mcs()


if __name__ == '__main__':
    # test_standard_case()
    test_arg_dict()
    # test_gui()



# distribution_auto_fitting
# Author: Yan Fu
# Date: 07/03/2019

import os
import warnings
from tkinter import filedialog, Tk, StringVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as st
from typing import Union
from scipy import interpolate


# Create models from data
def fit(data, bins:int=200, ax=None, distribution_list:Union[list, int]=None, suppress_print:bool=False):
    """
    Model data by finding best fit distribution to data
    :param data:
    :param bins:
    :param ax:
    :param distribution_list:
    :param suppress_print:
    :return list_fitted_distribution:
    :return list_fitted_distribution_params:
    :return list_fitted_distribution_sse:
    """

    # Get histogram of original data
    y, x = np.histogram(data, bins=bins, density=True)
    x = (x + np.roll(x, -1))[:-1] / 2.0
    # ax.plot(x, y)
    # ax.fill_between(x, y, color="black")

    def cprint(*args):
        if suppress_print:
            return 0
        else:
            print(*args)

    if ax:
        ax.hist(data, bins='auto', density=True, color='black')
        ax.set_xlim(min(*x) - (max(*x) - min(*x)) * 0.01, max(*x) + (max(*x) - min(*x)) * 0.01, auto=False)
        ax.set_ylim(min(*y) - (max(*y) - min(*y)) * 0.01, max(*y) + (max(*y) - min(*y)) * 0.01, auto=False)

    if distribution_list == 0 or distribution_list == None:
        # full list of available distribution functions
        distribution_list = [
            st.alpha,
            st.anglit,
            st.arcsine,
            st.beta,
            st.betaprime,
            st.bradford,
            st.burr,
            st.cauchy,
            st.chi,
            st.chi2,
            st.cosine,
            st.dgamma,
            st.dweibull,
            st.erlang,
            st.expon,
            st.exponnorm,
            st.exponweib,
            st.exponpow,
            st.f,
            st.fatiguelife,
            st.fisk,
            st.foldcauchy,
            st.foldnorm,
            st.frechet_r,
            st.frechet_l,
            st.genlogistic,
            st.genpareto,
            st.gennorm,
            st.genexpon,
            st.genextreme,
            st.gausshyper,
            st.gamma,
            st.gengamma,
            st.genhalflogistic,
            st.gilbrat,
            st.gompertz,
            st.gumbel_r,
            st.gumbel_l,
            st.halfcauchy,
            st.halflogistic,
            st.halfnorm,
            st.halfgennorm,
            st.hypsecant,
            st.invgamma,
            st.invgauss,
            st.invweibull,
            st.johnsonsb,
            st.johnsonsu,
            st.ksone,
            st.kstwobign,
            st.laplace,
            st.levy,
            st.levy_l,
            st.levy_stable,
            st.logistic,
            st.loggamma,
            st.loglaplace,
            st.lognorm,
            st.lomax,
            st.maxwell,
            st.mielke,
            st.nakagami,
            st.ncx2,
            st.ncf,
            st.nct,
            st.norm,
            st.pareto,
            st.pearson3,
            st.powerlaw,
            st.powerlognorm,
            st.powernorm,
            st.rdist,
            st.reciprocal,
            st.rayleigh,
            st.rice,
            st.recipinvgauss,
            st.semicircular,
            st.t,
            st.triang,
            st.truncexpon,
            st.truncnorm,
            st.tukeylambda,
            st.uniform,
            st.vonmises,
            st.vonmises_line,
            st.wald,
            st.weibull_min,
            st.weibull_max,
            st.wrapcauchy
        ]
    elif distribution_list == 1:
        # https://blog.cloudera.com/blog/2015/12/common-probability-distributions-the-data-scientists-crib-sheet/
        distribution_list = [
            st.uniform,
            st.bernoulli,
            st.hypergeom,
            st.binom,
            st.geom,
            st.gumbel_r,
            st.gumbel_l,
            st.nbinom,
            st.poisson,
            st.expon,
            st.lognorm,
            st.weibull_max,
            st.weibull_min,
            st.dweibull,
            st.norm,
            st.chi,
            st.gamma,
            st.beta,
        ]

    # Best holders
    best_distribution = st.norm
    best_params = (0.0, 1.0)
    best_sse = np.inf

    # Estimate distribution parameters from data

    cprint('{:20.20}{:60.60}'.format('Distribution', 'Loss (Residual sum of squares) and distribution parameters'))
    list_fitted_distribution = []
    list_fitted_distribution_sse = []
    list_fitted_distribution_params = []

    for distribution in distribution_list:

        # Try to fit the distribution
        try:
            # Ignore warnings from data that can't be fit
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore')

                # fit dist to data
                params = distribution.fit(data)

                # Separate parts of parameters
                arg = params[:-2]
                loc = params[-2]
                scale = params[-1]

                # Calculate fitted PDF and error with fit in distribution
                pdf = distribution.pdf(x, loc=loc, scale=scale, *arg)
                sse = np.sum(np.power(y - pdf, 2.0))

                # if axis pass in add to plot
                try:
                    if ax:
                        ax.plot(x, pdf, label=distribution.name)
                except Exception:
                    pass

                # identify if this distribution is better
                if best_sse > sse > 0:
                    best_distribution = distribution
                    best_params = params
                    best_sse = sse

                params_ = ["{:10.2f}".format(i) for i in params]
                cprint('{:20.20}{:10.5E} - [{}]'.format(distribution.name, sse, ", ".join(params_)))

                list_fitted_distribution.append(distribution)
                list_fitted_distribution_sse.append(sse)
                list_fitted_distribution_params.append(params)

        except Exception:
            pass

    return list_fitted_distribution, list_fitted_distribution_params, list_fitted_distribution_sse


def cdf_to_samples(cdf_x, cdf_y, n_samples=10000):
    cdf_func = interpolate.interp1d(cdf_y, cdf_x, kind='linear')
    cdf_pts = np.random.uniform(size=n_samples)
    return cdf_func(cdf_pts).flatten()


def pdf_to_samples(pdf_x, pdf_y, n_samples=10000):
    cdf_x = pdf_x
    cdf_y = np.cumsum(pdf_y)

    return cdf_to_samples(cdf_x, cdf_y, n_samples)


def auto_fit():
    # INPUTS

    # Note 1: *.csv must contain no headers and numbers only.

    # Define data type
    # 0 - single column containing sampled values
    # 1 - two columns with 1st column containing values and second column probability
    # 2 - two columns with 1st column containing values and second column cumulative probability
    data_type = input(
        "Define Data Type\n0 - single column containing sampled values\n1 - two columns with 1st column containing values and second column probability\n2 - two columns with 1st column containing values and second column cumulative probability\nEnter your value and press Enter: ")
    data_type = int(data_type)

    # Distribution type or list
    # list - will only fit to the provided list of distributions (scipy.stats.*)
    # 0 - fit to all available distributions
    # 1 - fit to common distribution types
    distribution_list = input(
        "Distribution Type\n0 - fit to all available distributions\n1 - fit to common distribution types\nEnter your value and press Enter: ")
    distribution_list = int(distribution_list)

    # Select data csv file
    root = Tk()
    root.withdraw()
    folder_path = StringVar()
    path_input_file_csv = filedialog.askopenfile(title='Select data csv file', filetypes=[('csv', ['.csv'])])
    folder_path.set(path_input_file_csv)
    root.update()

    dir_work = os.path.dirname(path_input_file_csv.name)
    print(dir_work)

    # Load data
    if data_type == 0:
        data_csv = pd.read_csv(path_input_file_csv, header=None, dtype=float)
        samples = data_csv.values.flatten()
    elif data_type == 1:
        data_csv = pd.read_csv(path_input_file_csv, header=None, dtype=float)
        samples = pdf_to_samples(pdf_x=data_csv[0].values.flatten(), pdf_y=data_csv[1].values.flatten(),
                                 n_samples=10000)
    elif data_type == 2:
        data_csv = pd.read_csv(path_input_file_csv, header=None, dtype=float)
        #     print(data_csv[1])
        samples = cdf_to_samples(cdf_x=data_csv[0].values.flatten(), cdf_y=data_csv[1].values.flatten(),
                                 n_samples=10000)
    else:
        samples = np.nan

    # FITTING DISTRIBUTIONS TO DATA

    fig_fitting, ax_fitting = plt.subplots(figsize=(3.94 * 2.5, 2.76 * 2.5))
    list_dist, list_params, list_sse = fit(samples, ax=ax_fitting, distribution_list=distribution_list)

    # FINDING THE BEST FIT

    list_dist = np.asarray(list_dist)[np.argsort(list_sse)]
    list_params = np.asarray(list_params)[np.argsort(list_sse)]
    list_sse = np.asarray(list_sse)[np.argsort(list_sse)]
    print('{:30.30}{:60.60}'.format('Distribution (sorted)',
                                    'Loss (Residual sum of squares) and distribution parameters'))
    for i, v in enumerate(list_dist):
        print('{:30.30}{:10.5E} - [{}]'.format(v.name, list_sse[i],
                                               ", ".join(["{:10.2f}".format(i) for j in list_params[i]])))

    dist_best = list_dist[0]
    params_best = list_params[0]

    # PRODUCE FIGURES

    print("Figures saved at", dir_work)

    ax_fitting.set_xlabel("Sample value")
    ax_fitting.set_ylabel("PDF")
    ax_fitting.legend().set_visible(True)
    fig_fitting.savefig(os.path.join(dir_work, "fitting.png"))

    cdf_x_sampled = np.sort(samples)
    cdf_y_sampled = np.linspace(0, 1, len(cdf_x_sampled))
    cdf_y_fitted = np.linspace(0, 1, 1000)
    cdf_x_fitted = dist_best.ppf(cdf_y_fitted, *params_best)

    fig_results, ax_results = plt.subplots(figsize=(3.94 * 2.5, 2.76 * 2.5))
    ax_results.hist(samples, bins='auto', density=True, color='black')
    ax_results.plot(cdf_x_sampled, cdf_y_sampled, label='Sampled', color='black')
    ax_results.plot(cdf_x_fitted, cdf_y_fitted, label='Fitted', color='red')
    ax_results.set_xlabel('Sample values')
    ax_results.set_ylabel('CDF')
    ax_results.legend().set_visible(True)
    fig_results.savefig(os.path.join(dir_work, "results.png"))

    input("Press any key to finish.")


if __name__ == '__main__':
    data_cdf_example = [[226.3874615,0],
        [229.4193217,0.001362959],
        [230.4470709,0.002016362],
        [231.4748201,0.002084425],
        [232.5539568,0.00381152],
        [233.4789311,0.002217148],
        [234.2497431,0.005579453],
        [235.3802672,0.008965581],
        [236.4594039,0.014003934],
        [237.2302158,0.019021869],
        [238.0524152,0.025698835],
        [238.9260021,0.032379205],
        [239.8509764,0.044029866],
        [240.8273381,0.057339559],
        [241.5981501,0.073946897],
        [242.5745118,0.087256590],
        [243.036999,0.1038435100],
        [243.8591984,0.125421139],
        [244.4758479,0.143673897],
        [245.04111,0.16357888100],
        [245.709147,0.1834906720],
        [246.3257965,0.201743430],
        [246.7368962,0.226605092],
        [247.3021583,0.251476964],
        [247.7646454,0.276342029],
        [248.2271326,0.304518353],
        [248.6896197,0.329383419],
        [249.2034943,0.355907516],
        [249.5632066,0.380765775],
        [250.1284687,0.403982018],
        [250.6937307,0.428853889],
        [251.1048304,0.458682439],
        [251.5673176,0.483547504],
        [252.0811922,0.506760344],
        [252.5950668,0.531628812],
        [253.1603289,0.561467571],
        [253.5714286,0.583017975],
        [254.1880781,0.611204508],
        [254.5991778,0.627788025],
        [255.2158273,0.659285816],
        [255.3186023,0.677504543],
        [255.626927,0.6891143660],
        [256.1921891,0.718953125],
        [256.5519013,0.735533238],
        [257.3741007,0.768700272],
        [257.7338129,0.778657868],
        [258.3504625,0.813466918],
        [258.6587873,0.833354887],
        [259.4809866,0.858243774],
        [260.2004111,0.873192080],
        [260.8684481,0.899726387],
        [261.2795478,0.909687387],
        [262.3586845,0.934593290],
        [262.8725591,0.944561097],
        [263.7461459,0.956208354],
        [264.5169579,0.967848805],
        [265.4933196,0.971224723],
        [266.1613566,0.977891481],
        [267.3946557,0.991218189],
        [268.5251799,0.994604317],
        [270.1181912,0.999676701],
        [275.2569373,1.000000000]]

    fit(cdf_to_samples(np.array(data_cdf_example)[:,0], np.array(data_cdf_example)[:,1]), distribution_list=1, suppress_print=False)

    # auto_fit()



"""
VERSION IDENTIFICATION RULES DOCUMENTED IN PEP 440 ARE FOLLOWED.

Version scheme
==============

Distributions are identified by a public version identifier which supports all defined version comparison operations

The version scheme is used both to describe the distribution version provided by a particular distribution archive, as
well as to place constraints on the version of dependencies needed in order to build or run the software.

Public version identifiers
--------------------------

The canonical public version identifiers MUST comply with the following scheme:

`[N!]N(.N)*[{a|b|rc}N][.postN][.devN]`

Public version identifiers MUST NOT include leading or trailing whitespace.

Public version identifiers MUST be unique within a given distribution.

See also Appendix B : Parsing version strings with regular expressions which provides a regular expression to check
strict conformance with the canonical format, as well as a more permissive regular expression accepting inputs that may
require subsequent normalization.

Public version identifiers are separated into up to five segments:

    - Epoch segment: N!
    - Release segment: N(.N)*
    - Pre-release segment: {a|b|rc}N
    - Post-release segment: .postN
    - Development release segment: .devN

"""

__version__ = '0.6.6'

if __name__ == '__main__':
    import re

    def is_canonical(version):
        return re.match(
            r'^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$',
            version) is not None


    assert is_canonical(__version__)



import numpy as np


def hydrocarbon_eurocode(
        time,
        temperature_initial
):
    time /= 1200.  # convert time unit from second to hour
    temperature_initial -= 273.15  # convert temperature from kelvin to celsius
    temperature = 1080 * (1 - 0.325 * np.exp(-0.167 * time) - 0.675 * np.exp(-2.5 * time)) + temperature_initial
    return temperature + 273.15



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

if __name__ == '__main__':
    import warnings
    import numpy as np
    from scipy.interpolate import interp1d
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs2.mcs2_calc import teq_main, teq_main_wrapper, mcs_out_post
    from sfeprapy.func.mcs_gen import main as gen

    warnings.filterwarnings('ignore')

    mcs = MCS()
    mcs.define_problem()
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(teq_main, teq_main_wrapper, mcs_out_post)
    mcs.run_mcs()



# -*- coding: utf-8 -*-
import cProfile


def profile_standard_case():
    import copy
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs0 import EXAMPLE_INPUT_DICT, EXAMPLE_CONFIG_DICT
    from sfeprapy.mcs0.mcs0_calc import teq_main, teq_main_wrapper, mcs_out_post
    from sfeprapy.func.mcs_gen import main as gen
    from scipy.interpolate import interp1d
    import numpy as np
    # increase the number of simulations so it gives sensible results
    mcs_input = copy.deepcopy(EXAMPLE_INPUT_DICT)
    mcs_config = copy.deepcopy(EXAMPLE_CONFIG_DICT)
    for k in list(mcs_input.keys()):
        mcs_input[k]['phi_teq'] = 1
        mcs_input[k]['n_simulations'] = 1000
        mcs_input[k]['timber_exposed_area'] = 0
        mcs_input[k].pop('beam_position_horizontal')
        mcs_input[k]['beam_position_horizontal:dist'] = 'uniform_'
        mcs_input[k]['beam_position_horizontal:ubound'] = mcs_input[k]['room_depth'] * 0.9
        mcs_input[k]['beam_position_horizontal:lbound'] = mcs_input[k]['room_depth'] * 0.6

    # increase the number of threads so it runs faster
    mcs_config['n_threads'] = 3
    mcs = MCS()
    mcs.define_problem(data=mcs_input, config=mcs_config)
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(teq_main, teq_main_wrapper, mcs_out_post)
    mcs.run_mcs()


if __name__ == '__main__':
    # Profile for the standard case input sfeprapy.mcs0.EXAMPLE_INPUT_CSV
    cProfile.run('profile_standard_case()', sort='cumtime')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


def save_figure(mcs_out, fp):
    import numpy as np
    import plotly.graph_objects as go
    import plotly.io

    def cdf_xy(v, xlim, bin_width=0.1):
        edges = np.arange(*xlim, bin_width)
        x = (edges[1:] + edges[:-1]) / 2
        y_pdf = np.histogram(v, edges)[0] / len(v)
        y_cdf = np.cumsum(y_pdf)
        return x, y_cdf

    fig_size = (3.5, 3.5)
    fig_x_limit = (0, 120)
    fig_y_limit = (0, 1)
    bin_width = 0.1

    dict_teq = dict()
    for case_name in set(mcs_out['case_name'].values):
        dict_teq[case_name] = np.asarray(
            mcs_out[mcs_out['case_name'] == case_name]['solver_time_equivalence_solved'].values, dtype=float) / 60.

    xlim = (0, np.max([np.max(v) for k, v in dict_teq.items()]) + bin_width)
    y_cdf = dict()
    for k, v in dict_teq.items():
        x, y_cdf[k] = cdf_xy(v, xlim)

    # Create random data with numpy

    fig = go.Figure()

    # Add traces
    for case_name, y in y_cdf.items():
        fig.add_trace(
            go.Scatter(
                x=x,
                y=y,
                mode='lines',
                name=case_name
            )
        )

    fig.update_layout(
        autosize=False,
        width=400,
        height=400,
        margin=dict(
            l=20,
            r=20,
            b=20,
            t=20,
            pad=0
        ),
        paper_bgcolor='White',
        plot_bgcolor='White',
        xaxis=dict(
            title="Equivalent of time exposure [min]",
            dtick=30,
            range=fig_x_limit,
            showline=True,
            linewidth=1,
            linecolor='black',
            mirror=True,
            visible=True,
            showgrid=True,
            gridcolor='Black',
            gridwidth=1,
            ticks='outside',
            zeroline=False,
        ),
        yaxis=dict(
            title='CDF',
            dtick=0.2,
            range=fig_y_limit,
            showline=True,
            linewidth=1,
            linecolor='black',
            mirror=True,
            visible=True,
            showgrid=True,
            gridcolor='Black',
            gridwidth=1,
            ticks='outside',
            zeroline=False,
        ),
        legend=dict(
            x=0.98,
            xanchor='right',
            y=0.02,
            yanchor='bottom',
            traceorder="normal",
            font=dict(
                family="sans-serif",
                size=12,
                color="black"
            ),
            bgcolor="White",
            bordercolor="Black",
            borderwidth=1,
        ),
    )
    plotly.io.write_html(fig, file=fp, auto_open=False)


if __name__ == '__main__':
    import os
    import sys
    import warnings
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs0.mcs0_calc import teq_main, teq_main_wrapper, mcs_out_post
    from sfeprapy.func.mcs_gen import main as gen

    warnings.filterwarnings('ignore')

    mcs_problem_definition = None
    n_threads = None
    fig = None
    if len(sys.argv) > 1:
        mcs_problem_definition = os.path.realpath(sys.argv[1])
    if len(sys.argv) > 2:
        for arg in sys.argv[2:]:
            if 'mp' in arg:
                n_threads = int(str(arg).replace('mp', ''))
            if 'fig' in arg:
                fig = True

    mcs = MCS()
    mcs.define_problem(mcs_problem_definition)
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(teq_main, teq_main_wrapper, mcs_out_post)

    if n_threads:
        mcs.config = dict(n_threads=n_threads) if mcs.config else mcs.config.update(dict(n_threads=n_threads))

    mcs.run_mcs()

    if fig:
        save_figure(mcs_out=mcs.mcs_out, fp=os.path.join(os.path.dirname(mcs_problem_definition), 'mcs.out.html'))



# -*- coding: utf-8 -*-

def __example_config_dict():
    y = dict(
    n_threads=2,
    )
    return y


def __example_input_dict():
    y = {
        "Standard Case 1": dict(
            case_name='Standard Case 1',
            n_simulations=100,
            probability_weight=0.5,
            fire_time_step=30,
            fire_time_duration=18000,

            fire_hrr_density=dict(
                dist='uniform_',
                lbound=0.25 - 0.001,
                ubound=0.25 + 0.001),
            fire_load_density=dict(
                dist='gumbel_r_',
                lbound=10,
                ubound=1500,
                mean=420,
                sd=126),
            fire_spread_speed=dict(
                dist='uniform_',
                lbound=0.0035,
                ubound=0.0190),
            fire_nft_limit=dict(
                dist='norm_',
                lbound=623.15,
                ubound=2023.15,
                mean=1323.15,
                sd=93),
            fire_combustion_efficiency=dict(
                dist='uniform_',
                lbound=0.8,
                ubound=1.0),
            window_open_fraction=dict(
                dist='lognorm_mod_',
                ubound=0.9999,
                lbound=0.0001,
                mean=0.2,
                sd=0.2),
            phi_teq=dict(
                dist='lognorm_',
                ubound=3,
                lbound=0.00001,
                mean=1,
                sd=0.25
            ),

            beam_cross_section_area=0.017,
            beam_position_horizontal=-1,
            beam_position_vertical=3.2,
            beam_rho=7850,
            fire_mode=3,
            fire_gamma_fi_q=1,
            fire_t_alpha=300,
            fire_tlim=0.333,
            protection_c=1700,
            protection_k=0.2,
            protection_protected_perimeter=2.14,
            protection_rho=800,
            room_breadth=16,
            room_depth=31.25,
            room_height=3.3,
            room_wall_thermal_inertia=720,
            solver_temperature_goal=893.15,
            solver_max_iter=20,
            solver_thickness_lbound=0.0001,
            solver_thickness_ubound=0.0500,
            solver_tol=1.,
            window_height=2.8,
            window_width=72,
            window_open_fraction_permanent=0,

            timber_exposed_area=10,
            timber_charring_rate=0.7,  # mm/min
            timber_hc=13.2,  # MJ/kg
            timber_density=400,  # [kg/m3]
            timber_solver_ilim=20,
            timber_solver_tol=1
        ),

        "Standard Case 2": dict(
            case_name='Standard Case 2',
            n_simulations=100,
            probability_weight=0.5,
            fire_time_step=30,
            fire_time_duration=18000,

            fire_hrr_density=dict(
                dist='uniform_',
                lbound=0.25 - 0.001,
                ubound=0.25 + 0.001),
            fire_load_density=dict(
                dist='gumbel_r_',
                lbound=10,
                ubound=1500,
                mean=420,
                sd=126),
            fire_spread_speed=dict(
                dist='uniform_',
                lbound=0.0035,
                ubound=0.0190),
            fire_nft_limit=dict(
                dist='norm_',
                lbound=623.15,
                ubound=2023.15,
                mean=1323.15,
                sd=93),
            fire_combustion_efficiency=dict(
                dist='uniform_',
                lbound=0.8,
                ubound=1.0),
            window_open_fraction=dict(
                dist='lognorm_mod_',
                ubound=0.9999,
                lbound=0.0001,
                mean=0.2,
                sd=0.2),
            phi_teq=dict(
                dist='lognorm_',
                ubound=3,
                lbound=0.00001,
                mean=1,
                sd=0.25
            ),

            beam_cross_section_area=0.017,
            beam_position_horizontal=-1,
            beam_position_vertical=3.2,
            beam_rho=7850,
            fire_mode=3,
            fire_gamma_fi_q=1,
            fire_t_alpha=300,
            fire_tlim=0.333,
            protection_c=1700,
            protection_k=0.2,
            protection_protected_perimeter=2.14,
            protection_rho=800,
            room_breadth=16,
            room_depth=31.25,
            room_height=3.3,
            room_wall_thermal_inertia=720,
            solver_temperature_goal=893.15,
            solver_max_iter=20,
            solver_thickness_lbound=0.0001,
            solver_thickness_ubound=0.0500,
            solver_tol=1.,

            window_height=2.8,
            window_width=72,
            window_open_fraction_permanent=0,

            timber_exposed_area=10,
            timber_charring_rate=0.7,  # mm/min
            timber_hc=13.2,  # MJ/kg
            timber_density=400,  # [kg/m3]
            timber_solver_ilim=20,
            timber_solver_tol=1
        ),
    }
    return y


def __example_input_csv(x: dict):
    from sfeprapy.func.mcs_gen import dict_flatten
    import pandas as pd
    y = {k: dict_flatten(v) for k, v in x.items()}
    y = pd.DataFrame.from_dict(y, orient='columns')
    y.index.name = 'PARAMETERS'
    y = y.to_csv(index=True)
    return y


EXAMPLE_CONFIG_DICT = __example_config_dict()
EXAMPLE_INPUT_DICT = __example_input_dict()
EXAMPLE_INPUT_CSV = __example_input_csv(__example_input_dict())

if __name__ == '__main__':
    print(EXAMPLE_CONFIG_DICT, '\n')
    print(EXAMPLE_INPUT_DICT, '\n')
    print(EXAMPLE_INPUT_CSV, '\n')



# -*- coding: utf-8 -*-


def test_standard_case():
    import copy
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs0 import EXAMPLE_INPUT_DICT, EXAMPLE_CONFIG_DICT
    from sfeprapy.mcs0.mcs0_calc import teq_main, teq_main_wrapper, mcs_out_post
    from sfeprapy.func.mcs_gen import main as gen
    from scipy.interpolate import interp1d
    import numpy as np
    # increase the number of simulations so it gives sensible results
    mcs_input = copy.deepcopy(EXAMPLE_INPUT_DICT)
    mcs_config = copy.deepcopy(EXAMPLE_CONFIG_DICT)
    for k in list(mcs_input.keys()):
        mcs_input[k]['phi_teq'] = 1
        mcs_input[k]['n_simulations'] = 1000
        mcs_input[k]['timber_exposed_area'] = 0
        mcs_input[k].pop('beam_position_horizontal')
        mcs_input[k]['beam_position_horizontal:dist'] = 'uniform_'
        mcs_input[k]['beam_position_horizontal:ubound'] = mcs_input[k]['room_depth'] * 0.9
        mcs_input[k]['beam_position_horizontal:lbound'] = mcs_input[k]['room_depth'] * 0.6

    # increase the number of threads so it runs faster
    mcs_config['n_threads'] = 3
    mcs = MCS()
    mcs.define_problem(data=mcs_input, config=mcs_config)
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(teq_main, teq_main_wrapper, mcs_out_post)
    mcs.run_mcs()
    mcs_out = mcs.mcs_out
    teq = mcs_out['solver_time_equivalence_solved'] / 60.
    hist, edges = np.histogram(teq, bins=np.arange(0, 181, 0.5))
    x, y = (edges[:-1] + edges[1:]) / 2, np.cumsum(hist / np.sum(hist))
    teq_at_80_percentile = interp1d(y, x)(0.8)
    print(teq_at_80_percentile)
    target, target_tol = 60, 1
    assert target - target_tol < teq_at_80_percentile < target + target_tol


# test non-gui version
def test_arg():
    import copy
    import numpy as np
    from scipy.interpolate import interp1d
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs0 import EXAMPLE_INPUT_DICT, EXAMPLE_CONFIG_DICT
    from sfeprapy.mcs0.mcs0_calc import teq_main, teq_main_wrapper, mcs_out_post
    from sfeprapy.func.mcs_gen import main as gen
    mcs_input = copy.deepcopy(EXAMPLE_INPUT_DICT)
    mcs_config = copy.deepcopy(EXAMPLE_CONFIG_DICT)
    for k in list(mcs_input.keys()):
        mcs_input[k].pop('phi_teq')
        mcs_input[k]['phi_teq:dist'] = 'lognorm_'
        mcs_input[k]['phi_teq:ubound'] = 3
        mcs_input[k]['phi_teq:lbound'] = 0.00001
        mcs_input[k]['phi_teq:mean'] = 1
        mcs_input[k]['phi_teq:sd'] = 0.25
        mcs_input[k]['n_simulations'] = 1000
        mcs_input[k]['timber_exposed_area'] = 0

    mcs_config['n_threads'] = 3
    mcs = MCS()
    mcs.define_problem(data=mcs_input, config=mcs_config)
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(teq_main, teq_main_wrapper, mcs_out_post)
    mcs.run_mcs()
    mcs_out = mcs.mcs_out
    teq = mcs_out['solver_time_equivalence_solved'] / 60.
    hist, edges = np.histogram(teq, bins=np.arange(0, 181, 0.5))
    x, y = (edges[:-1] + edges[1:]) / 2, np.cumsum(hist / np.sum(hist))
    teq_at_80_percentile = interp1d(y, x)(0.8)
    print(teq_at_80_percentile)


if __name__ == '__main__':
    test_standard_case()
    test_arg()



import numpy as np


def fire(
        time,
        temperature_ambient
):
    time /= 1200.  # convert from seconds to hours
    temperature_ambient -= 273.15  # convert temperature from kelvin to celcius
    temperature = 750 * (1 - np.exp(-3.79553 * np.sqrt(time))) + 170.41 * np.sqrt(time) + temperature_ambient
    return temperature + 273.15  # convert from celsius to kelvin (SI unit)



import numpy as np


def fire(t, A_t, A_f, A_v, h_eq, q_fd, lambda_, rho, c, t_lim, temperature_initial=293.15):
    """Function Description: (SI UNITS ONLY)
    This function calculates the time-temperature curve according to Eurocode 1 part 1-2, Appendix A.
    :param t: numpy.ndarray, [s], time evolution.
    :param A_t: float, [m2], total surface area (including openings).
    :param A_f: float, [m2], floor area.
    :param A_v: float, [m2], opening area.
    :param h_eq: float, [m2], opening height.
    :param q_fd: float, [J/m2], fuel density.
    :param lambda_: float, [K/kg/m], lining thermal conductivity.
    :param rho: float, [kg/m3], lining density.
    :param c: float, [J/K/kg], lining thermal capacity.
    :param t_lim: float, [s], limiting time for the fire.
    :return T_g: numpy.ndarray, [s], temperature evolution.
    """
    # Reference: Eurocode 1991-1-2; Jean-Marc Franssen, Paulo Vila Real (2010) - Fire Design of Steel Structures

    # UNITS: SI -> Equations
    q_fd /= 1e6  # [J/m2] -> [MJ/m2]
    t_lim /= 3600  # [s] -> [hr]
    t = t / 3600  # [s] -> [hr]
    temperature_initial -= 273.15  # [K] -> [C]

    # ACQUIRING REQUIRED VARIABLES
    # t = np.arange(time_start, time_end, time_step, dtype=float)

    b = (lambda_ * rho * c) ** 0.5
    O = A_v * h_eq**0.5 / A_t
    q_td = q_fd * A_f / A_t
    Gamma = ((O/0.04)/(b/1160))**2

    t_max = 0.0002*q_td/O

    # check criteria
    # todo: raise warnings to higher level
    # if not 50 <= q_td <= 1000:
    #     if is_cap_q_td:
    #         msg = "q_td ({:4.1f}) EXCEEDED [50, 1000] AND IS CAPPED.".format(q_td, is_cap_q_td)
    #     else:
    #         msg = "q_td ({:4.1f}) EXCEEDED [50, 1000] AND IS UNCAPPED.".format(q_td)
    #     warnings.warn(msg)

    # CALCULATION
    def _temperature_heating(t_star, temperature_initial):
        # eq. 3.12
        T_g = 1325 * (1 - 0.324 * np.exp(-0.2 * t_star) - 0.204 * np.exp(-1.7*t_star) - 0.472*np.exp(-19*t_star))
        T_g += temperature_initial
        return T_g

    def _temperature_cooling_vent(t_star_max, T_max, t_star):  # ventilation controlled
        # eq. 3.16
        if t_star_max <= 0.5:
            T_g = T_max - 625 * (t_star - t_star_max)
        elif 0.5 < t_star_max < 2.0:
            T_g = T_max - 250 * (3 - t_star_max) * (t_star - t_star_max)
        elif 2.0 <= t_star_max:
            T_g = T_max - 250 * (t_star - t_star_max)
        else: T_g = np.nan
        return T_g

    def _temperature_cooling_fuel(t_star_max, T_max, t_star, Gamma, t_lim):  # fuel controlled
        # eq. 3.22
        if t_star_max <= 0.5:
            T_g = T_max - 625 * (t_star - Gamma * t_lim)
        elif 0.5 < t_star_max < 2.0:
            T_g = T_max - 250 * (3 - t_star_max) * (t_star - Gamma * t_lim)
        elif 2.0 <= t_star_max:
            T_g = T_max - 250 * (t_star - Gamma * t_lim)
        else: T_g = np.nan
        return T_g

    def _variables(t, Gamma, t_max):
        t_star = Gamma * t
        t_star_max = Gamma * t_max
        return t_star, t_star_max

    def _variables_2(t, t_lim, q_td, b, O):
        O_lim = 0.0001 * q_td / t_lim
        Gamma_lim = ((O_lim/0.04)/(b/1160))**2

        if O > 0.04 and q_td < 75 and b < 1160:
            k = 1 + ((O-0.04)/(0.04)) * ((q_td-75)/(75)) * ((1160-b)/(1160))
            Gamma_lim *= k

        t_star_ = Gamma_lim * t
        t_star_max_ = Gamma_lim * t_lim
        return t_star_, t_star_max_

    t_star, t_star_max = _variables(t, Gamma, t_max)

    if t_max >= t_lim:  # ventilation controlled fire
        T_max = _temperature_heating(t_star_max, temperature_initial)
        T_heating_g = _temperature_heating(Gamma * t, temperature_initial)
        T_cooling_g = _temperature_cooling_vent(t_star_max, T_max, t_star)
        fire_type = "ventilation controlled"
    else:  # fuel controlled fire
        t_star_, t_star_max_ = _variables_2(t, t_lim, q_td, b, O)
        T_max = _temperature_heating(t_star_max_, temperature_initial)
        T_heating_g = _temperature_heating(t_star_, temperature_initial)
        T_cooling_g = _temperature_cooling_fuel(t_star_max, T_max, t_star, Gamma, t_lim)
        fire_type = "fuel controlled"

    T_g = np.minimum(T_heating_g, T_cooling_g)
    T_g[T_g < temperature_initial] = temperature_initial

    # UNITS: Eq. -> SI
    t *= 3600
    T_g += 273.15

    return T_g


if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set_style('ticks')
    sns.set_context(context='paper')

    fig, ax = plt.subplots(figsize=(3.94, 2.76))

    # define geometry
    w = 16
    l = 32
    h = 3

    h_v = 2

    oo = (0.02, 0.04, 0.06, 0.10, 0.14, 0.20)  # desired opening factor

    def opening_factor_2_window_width(of, w, l, h, h_v):
        # opening factor = Av * sqrt(hv) / At
        # Av = h_v * w_v
        # of * At / sqrt(h_v) = h_v * w_v
        # w_v = of * At / sqrt(h_v) / h_v
        return of * (2 * (w * l + l * h + h * w)) / h_v ** 0.5 / h_v

    w_v_ = [opening_factor_2_window_width(o, w, l, h, h_v) for o in oo]

    print(w_v_)

    # Calculate fire curves
    for w_v in w_v_:
        x = np.arange(0, 5 * 60 * 60, 1)
        y = fire(
            t=x,
            A_v=h_v * w_v,
            A_t=2*(w*h+h*l+l*w),
            A_f=w*l,
            h_eq=h_v,
            q_fd=600e6,
            lambda_=1,
            rho=1,
            c=720**2,
            t_lim=20 * 60,
            temperature_initial=293.15
        )
        ax.plot(x / 60, y-273.15, label="Opening Factor {:.2f}".format((h_v * w_v) * h_v**0.5 / (2*(w*h+h*l+l*w))))

    ax.set_xlabel('Time [minute]')
    ax.set_ylabel('Temperature [$^{\circ}C$]')

    ax.legend(loc=4).set_visible(False)

    # ax.grid(color='k', linestyle='--')
    plt.tight_layout()
    plt.show()

    fig.add_subplot()



# -*- coding: utf-8 -*-
import os
from tkinter import Tk, StringVar, filedialog
import pandas as pd
import numpy as np
from scipy.interpolate import interp1d


def tk_select_path():

    root = Tk()
    root.withdraw()
    folder_path = StringVar()

    path_folder = filedialog.askdirectory(title='Select Input Files')
    folder_path.set(path_folder)
    root.update()

    return path_folder


def get_list_file_path_with_suffix(path_folder, suffix):

    list_path_out_csv_all = []

    for f in os.listdir(path_folder):
        if f.endswith(suffix):
            list_path_out_csv_all.append(os.path.join(path_folder, f))

    return list_path_out_csv_all


def read_csvs_to_dfs(list_path_csv_file):

    list_pd_out_csv = []

    for path_out_csv in list_path_csv_file:
        df_ = pd.read_csv(path_out_csv, index_col='INDEX', dtype=np.float64)

        list_pd_out_csv.append(df_)

    return list_pd_out_csv


def get_teq_for_fr(list_df_out, list_fire_resistance):

    list_teq = []

    for df in list_df_out:

        # obtain x and y values for plot

        mask_teq_sort = np.asarray(df["TIME EQUIVALENCE [s]"].values).argsort()

        x = df["TIME EQUIVALENCE [s]"].values[mask_teq_sort] / 60.  # to minutes
        y = np.arange(1, len(x) + 1) / len(x)

        f_interp = interp1d(x, y, bounds_error=False, fill_value=(min(y), max(y)))

        list_teq.append(f_interp(list_fire_resistance))

    return list_teq


def main():

    path_folder = tk_select_path()

    list_path_files = get_list_file_path_with_suffix(path_folder, '_out.csv')

    list_df = read_csvs_to_dfs(list_path_files)

    list_teq = get_teq_for_fr(list_df, [30, 60, 90, 120, 150, 180])

    fmt = '{:<20.18},{:10.6f},{:10.6f},{:10.6f},{:10.6f},{:10.6f},{:10.6f}'

    for i, path_file in enumerate(list_path_files):

        file_name = os.path.basename(path_file).replace('_out.csv', '')

        print(fmt.format(file_name, *list_teq[i]))


if __name__ == '__main__':

    main()



import copy
import json
import os
import time
import warnings
from tkinter import filedialog, Tk, StringVar
from typing import Union, Callable

import pandas as pd
from tqdm import tqdm


class MCS:
    DEFAULT_TEMP_FOLDER_NAME = 'mcs.out'
    DEFAULT_MCS_OUTPUT_FILE_NAME = 'mcs.out.csv'
    DEFAULT_CONFIG_FILE_NAME = 'config.json'
    DEFAULT_CONFIG = dict(n_threads=1)

    def __init__(self):
        self._path_wd = None

        self._dict_master_input = None
        self._dict_config = None

        self._func_mcs_gen = None
        self._func_mcs_calc = None
        self._func_mcs_calc_mp = None
        self._func_mcs_post_1 = None

        self._df_mcs_out = None

    @property
    def path_wd(self) -> str:
        return self._path_wd

    @property
    def input(self) -> dict:
        return self._dict_master_input

    @property
    def config(self) -> dict:
        return self._dict_config

    @property
    def func_mcs_gen(self) -> Callable:
        return self._func_mcs_gen

    @property
    def func_mcs_calc(self) -> Callable:
        return self._func_mcs_calc

    @property
    def func_mcs_calc_mp(self) -> Callable:
        return self._func_mcs_calc_mp

    @property
    def func_mcs_post(self) -> Callable:
        return self._func_mcs_post_1

    @property
    def mcs_out(self) -> pd.DataFrame:
        return self._df_mcs_out

    @path_wd.setter
    def path_wd(self, p_):
        assert os.path.isdir(p_)
        self._path_wd = p_

    @input.setter
    def input(self, dict_: dict):
        self._dict_master_input = dict_

    @config.setter
    def config(self, dict_config: dict):
        self._dict_config = dict_config

    @func_mcs_gen.setter
    def func_mcs_gen(self, func_mcs_gen: Callable):
        self._func_mcs_gen = func_mcs_gen

    @func_mcs_calc.setter
    def func_mcs_calc(self, func_mcs_calc: Callable):
        self._func_mcs_calc = func_mcs_calc

    @func_mcs_calc_mp.setter
    def func_mcs_calc_mp(self, func_mcs_calc_mp: Callable):
        self._func_mcs_calc_mp = func_mcs_calc_mp

    @func_mcs_post.setter
    def func_mcs_post(self, func_mcs_post: Callable):
        self._func_mcs_post_1 = func_mcs_post

    @mcs_out.setter
    def mcs_out(self, df_out: pd.DataFrame):
        self._df_mcs_out = df_out

    def define_problem(self, data: Union[str, pd.DataFrame, dict] = None, config: dict = None, path_wd: str = None):

        # to get problem definition: try to parse from csv/xls/xlsx

        if data is None:
            fp = os.path.realpath(self._get_file_path_gui())
            self.path_wd = os.path.dirname(fp)
            if fp.endswith('.xlsx') or fp.endswith('.xls'):
                self.input = pd.read_excel(fp).set_index('PARAMETERS').to_dict()
            elif fp.endswith('.csv'):
                self.input = pd.read_csv(fp).set_index('PARAMETERS').to_dict()
            else:
                raise ValueError('Unknown input file format.')
        elif isinstance(data, str):
            fp = data
            self.path_wd = os.path.dirname(fp)
            if fp.endswith('.xlsx') or fp.endswith('.xls'):
                self.input = pd.read_excel(fp).set_index('PARAMETERS').to_dict()
            elif fp.endswith('.csv'):
                self.input = pd.read_csv(fp).set_index('PARAMETERS').to_dict()
            else:
                raise ValueError('Unknown input file format.')
        elif isinstance(data, pd.DataFrame):
            self.input = data.to_dict()
        elif isinstance(data, dict):
            self.input = data
        else:
            raise TypeError("Unknown input data type.")

        # to get configuration: try to parse from cwd if there is any, otherwise chose default values

        if config:  # first to check whether config is provided as parameter
            self.config = config
        else:  # otherwise
            try:
                with open(os.path.join(self.path_wd, self.DEFAULT_CONFIG_FILE_NAME), 'r') as f:
                    self.config = json.load(f)
            except (TypeError, FileNotFoundError):
                self.config = self.DEFAULT_CONFIG

    def define_stochastic_parameter_generator(self, func):
        self.func_mcs_gen = func

    def define_calculation_routine(self, func_mcs, func_mcs_mp=None, func_mcs_out_post=None):
        self.func_mcs_calc = func_mcs
        self.func_mcs_calc_mp = func_mcs_mp
        self.func_mcs_post = func_mcs_out_post

    def run_mcs(self):
        # Check whether required parameters are defined
        err_msg = list()
        if self._dict_master_input is None:
            err_msg.append('Problem definition is not defined.')
        if self._func_mcs_calc is None:
            err_msg.append('Monte Carlo Simulation calculation routine is not defined.')
        if self._func_mcs_gen is None:
            err_msg.append('Monte Carlo Simulation stochastic parameter generator is not defined.')
        if len(err_msg) > 0:
            raise ValueError(r'\n'.join(err_msg))

        # Prepare mcs parameter inputs
        x1 = self.input
        for case_name in list(x1.keys()):
            for param_name in list(x1[case_name].keys()):

                if ':' in param_name:
                    param_name_parent, param_name_sibling = param_name.split(':')
                    if param_name_parent not in x1[case_name]:
                        x1[case_name][param_name_parent] = dict()
                    x1[case_name][param_name_parent][param_name_sibling] = x1[case_name].pop(param_name)
        # to convert all "string numbers" to float data type
        for case_name in x1.keys():
            for i, v in x1[case_name].items():
                if isinstance(v, dict):
                    for ii, vv in v.items():
                        try:
                            x1[case_name][i][ii] = float(vv)
                        except:
                            pass
                else:
                    try:
                        x1[case_name][i] = float(v)
                    except:
                        pass

        # Generate mcs parameter samples
        x2 = {k: self.func_mcs_gen(v, int(v['n_simulations'])) for k, v in x1.items()}

        # Run mcs simulation
        x3 = dict()
        for k, v in x2.items():

            x3_ = self._mcs_mp(self.func_mcs_calc, self.func_mcs_calc_mp, x=v, n_threads=self.config['n_threads'])
            if self.func_mcs_post:
                x3_ = self.func_mcs_post(x3_)
            x3[k] = copy.copy(x3_)

            if self.path_wd:
                if not os.path.exists(os.path.join(self.path_wd, self.DEFAULT_TEMP_FOLDER_NAME)):
                    os.makedirs(os.path.join(self.path_wd, self.DEFAULT_TEMP_FOLDER_NAME))
                x3_.to_csv(os.path.join(os.path.join(self.path_wd, self.DEFAULT_TEMP_FOLDER_NAME), f'{k}.csv'))

        self.mcs_out = pd.concat([v for v in x3.values()])

        if self.path_wd:
            self.mcs_out.to_csv(os.path.join(self.path_wd, self.DEFAULT_MCS_OUTPUT_FILE_NAME), index=False)

    @staticmethod
    def _mcs_mp(func, func_mp, x: pd.DataFrame, n_threads: int) -> pd.DataFrame:
        list_mcs_in = x.to_dict(orient='records')

        time.sleep(0.5)  # to avoid clashes between the prints and progress bar
        print('{:<24.24}: {}'.format("CASE", list_mcs_in[0]['case_name']))
        print('{:<24.24}: {}'.format("NO. OF THREADS", n_threads))
        print('{:<24.24}: {}'.format("NO. OF SIMULATIONS", len(x.index)))
        time.sleep(0.5)  # to avoid clashes between the prints and progress bar

        if n_threads == 1 or func_mp is None:
            mcs_out = list()
            for i in tqdm(list_mcs_in, ncols=60):
                mcs_out.append(func(**i))
        else:
            import multiprocessing as mp
            m, p = mp.Manager(), mp.Pool(n_threads, maxtasksperchild=1000)
            q = m.Queue()
            jobs = p.map_async(func_mp, [(dict_, q) for dict_ in list_mcs_in])
            n_simulations = len(list_mcs_in)
            with tqdm(total=n_simulations, ncols=60) as pbar:
                while True:
                    if jobs.ready():
                        if n_simulations > pbar.n:
                            pbar.update(n_simulations - pbar.n)
                        break
                    else:
                        if q.qsize() - pbar.n > 0:
                            pbar.update(q.qsize() - pbar.n)
                        time.sleep(1)
                p.close()
                p.join()
                mcs_out = jobs.get()
                time.sleep(0.5)

        # clean and convert results to dataframe and return
        df_mcs_out = pd.DataFrame(mcs_out)
        df_mcs_out.sort_values('solver_time_equivalence_solved', inplace=True)  # sort base on time equivalence
        return df_mcs_out

    @staticmethod
    def _get_file_path_gui():
        root = Tk()
        root.withdraw()
        folder_path = StringVar()

        path_input_file_csv = filedialog.askopenfile(title='Select Input File',
                                                     filetypes=[('MCS IN', ['.csv', '.xlsx'])])
        folder_path.set(path_input_file_csv)
        root.update()

        try:
            path_input_file_csv = os.path.realpath(path_input_file_csv.name)
            return path_input_file_csv
        except AttributeError:
            raise FileNotFoundError('File not found.')


def test():

    # test gui version
    def test_gui():
        from sfeprapy.mcs0.mcs0_calc import teq_main as calc
        from sfeprapy.mcs0.mcs0_calc import teq_main_wrapper as calc_mp
        from sfeprapy.func.mcs_gen import main as gen
        mcs = MCS()
        mcs.define_problem()
        mcs.define_stochastic_parameter_generator(gen)
        mcs.define_calculation_routine(calc, calc_mp)
        mcs.run_mcs()

    # test non-gui version
    def test_arg_dict():
        import sfeprapy.mcs0 as mcs0
        from sfeprapy.mcs0.mcs0_calc import teq_main as calc
        from sfeprapy.mcs0.mcs0_calc import teq_main_wrapper as calc_mp
        from sfeprapy.func.mcs_gen import main as gen
        mcs = MCS()
        mcs.define_problem(data=mcs0.EXAMPLE_INPUT_DICT, config=mcs0.EXAMPLE_CONFIG_DICT)
        mcs.define_stochastic_parameter_generator(gen)
        mcs.define_calculation_routine(calc, calc_mp)
        mcs.run_mcs()

    test_gui()
    test_arg_dict()


if __name__ == '__main__':
    warnings.filterwarnings('ignore')
    test()



# __author__ = "Danny Hopkin"
# __email__= "Danny.Hopkin@OFRconsultants.com"
# __date__= "02-05-19"

import numpy as np
from scipy.stats import gamma
from scipy.stats import lognorm
from scipy.stats import norm


## inverse CDF function ##

def Finv_Normal(r, m, s):
    return norm.ppf(r, m, s)


def Finv_Lognormal(r, m, s):
    sln, mln = p_Lognormal(m, s)

    return lognorm.ppf(r, sln, 0, np.exp(mln))

## parameter calculation ##

def p_Lognormal(m, s):
    cov = s / m;

    sln = np.sqrt(np.log(1 + cov ** 2))
    mln = np.log(m) - 1 / 2 * sln ** 2

    return sln, mln

def p_Gamma(m, s):
    # --------------------------------------------------------------------------------------------------------
    # X ~ Gamma distribution
    # Input:m,s
    # Intermediate calc: k,lambda cfr. W&S p.5.8
    # Output: scale, loc cfr. scipy
    # --------------------------------------------------------------------------------------------------------
    # parameters Gamma W&S
    l = m / s ** 2
    k = m * l
    # parameters Gamma scipy
    a = k
    scale = 1 / l
    return a, scale


def ky_req_calc(n_samples):
    ########### Testing #####################

    # Starting section utilisation

    u = 1.0  # Section utilisation between 0 and 1.0 based on ULS ambient design

    # Live load parameters

    NomQ = 3  # Nominal instantaneous imposed load (Q) [kN/sq.m]
    Vimp = 0.95  # Instantaneous imposed load COV
    Mimp = 0.2 * NomQ  # Mean instantaneous imposed load [kN/sq.m]
    Simp = Vimp * Mimp  # St. Dev of instantaneous imposed load [kN/sq.m]
    GammaQ = 1.5  # ULS ambient partial factor on Q
    a, scale = p_Gamma(Mimp, Simp)  # Get distribution parameters

    # Permanent load parameters

    NomP = 3  # Nominal permanent load (G) [kN/sq.m]
    Vper = 0.1  # Permanent load (G) COV
    Mper = 1.0 * NomP  # Mean permanent load (G) [kN/sq.m]
    GammaG = 1.35  # ULS ambient partial factor on G
    Sper = Vper * Mper  # St. Dev of permanent load [kN/sq.m]

    # Model uncertainty parameters

    Mk = 1.0  # Mean total load model uncertainty [-]
    Vk = 0.1  # COV for total model uncertainty [-]
    Sk = Vk * Mk  # Standard dev of total model uncertainty [-]

    # Generate random variables

    samples = n_samples  # Number of samples for combined load distribution

    xT = np.linspace(0, 1, samples)  # probability array for plotting
    xQ = np.random.rand(samples)  # random number array for Q
    xG = np.random.rand(samples)  # random number array for G
    xK = np.random.rand(samples)  # random number array for KE

    Qp = gamma.ppf(xQ, a, loc=0, scale=scale)  # Calculate imposed load dist
    Gp = norm.ppf(xG, Mper, Sper)  # Calculate permanent load dist
    Kp = Finv_Lognormal(xK, Mk, Sk)  # Calculate model uncertainty factor
    Tp = Kp * (Gp + Qp)  # Total load model
    Tpscale = (u * Tp) / ((GammaG * NomP) + (GammaQ * NomQ))  # Normalisation relative to ambient ULS
    # Tpscale = np.sort(Tpscale)  # Sort for CDF
    # weights = np.ones_like(Tpscale) / float(len(Tpscale))  # Calculate weights for histogram

    return Tpscale

    # Plot results

    # plt.figure(1)
    # plt.subplot(211)
    # plt.xlabel('ky,req [-]')
    # plt.ylabel('CDF [-]')
    # plt.xlim([0, 1.0])
    # plt.ylim([0, 1.0])
    # plt.plot(Tpscale, xT, "y-", label=(r'$\alpha=29, \beta=3$'))
    # plt.subplot(212)
    # plt.xlabel('ky,req [-]')
    # plt.ylabel('PDF [-]')
    # plt.xlim([0, 1.0])
    # plt.hist(Tpscale, bins=200, weights=weights)
    # plt.show()

    # mcs_out = np.array(Tpscale)
    # np.savetxt('test.csv', mcs_out)


if __name__ == '__main__':
    res = ky_req_calc(5000)
    np.savetxt('kyr.csv', res)


# -*- coding: utf-8 -*-
import numpy as np
from scipy.interpolate import interp1d
import warnings


def protected_steel_bs13381(
        time, temperature_ambient,
        lambda_pt_T, d_p, A_p, V, c_a_T, rho_a_T,
        time_ubound=10800, time_step_lbound=30.
):
    """
    :param time:                {ndarray}   [s]         Time array incorporating temperature_ambient, forming a time-temperature curve
    :param temperature_ambient: {ndarray}   [s]         Temperature array incorporating time, forming a time-temperature curve
    :param lambda_pt_T:         {ky2T}      [W/m/K]     Thermal conductivity of the protective material (temperature dependent)
    :param d_p:                 {float}     [m]         Dry film thickness of reactive product
    :param A_p:                 {float}     [m]         Perimeter of protected
    :param V:                   {float}     [m2]        Area of steel cross section
    :param c_a_T:               {ky2T}      [J/kg/K]    Specific heat capacity of steel (temperature dependent)
    :param rho_a_T:             {ky2T}      [kg/m3]     Density of steel (temperature dependent)
    :param time_ubound:         {float}     [s]         The calculation time
    :param time_step_lbound:    {float}     [s]         The user-defined minimum time step, however, a smaller time step will be used with E.2
    :return:
    """

    # DEFINE FUNCTIONS: There are several functions defined in BS EN 13381-8:2013, ANNEX E, E.3. They are defined in
    # advance for simplification.
    # [BS EN 13381-8:2013, ANNEX E, Equation E.1]
    def _theta_at(lambda_pt, d_p, A_p, V, c_a, rho_a, theta_t, theta_at, dt):
        return (lambda_pt/d_p) * (A_p/V) * (1/c_a/rho_a) * (theta_t-theta_at) * dt

    # [BS EN 13381-8:2013, ANNEX E, Equation E.2]
    def _dt(c_a, rho_a, lambda_pt, d_p, A_p, V):
        return 0.8 * (c_a * rho_a / lambda_pt * d_p) * (V / A_p)

    # [BS EN 13381-8:2013, ANNEX E, Equation E.3]
    def _lambda_pt(d_p, V, A_p, c_a, rho_a, theta_t, theta_at, dt, d_theta_at):
        return d_p * (V / A_p) * c_a * rho_a * (1 / ((theta_t-theta_at)*dt)) * d_theta_at

    # [BS EN 13381-8:2013, ANNEX E, Equation E.4]
    def _theta_pt(theta_t, theta_t_, theta_at, theta_at_):
        return ((theta_t_+theta_t)/2. + (theta_at_+theta_at)/2.) / 2.

    # [BS EN 13381-8:2013, ANNEX E, Equation E.5]
    def _d_theta_at(c_a, rho_a, lambda_ave, d_p, A_p, V, theta_t, theta_at, dt):
        return (1/(c_a+rho_a)) * (lambda_ave/d_p) * (A_p / V) * (theta_t - theta_at) * dt

    # [BS EN 13381-8:2013, ANNEX E, Equation E.6]
    def _lambda_char(lambda_ave, K, sigma):
        return lambda_ave + K * sigma

    # Note: only equation E.1 and E.2 are currently being used as it is assumed the conductivity of protection layer
    #       is known.

    _temperature_gas = interp1d(time, temperature_ambient)

    # Instantiate output containers
    time_ = [0]
    time_rate = [0]
    temperature_steel = [temperature_ambient[0]]
    temperature_rate_steel = [0]
    conductivity_protection = [np.nan]

    # Start iterative calculation
    t = 0
    i = 0
    while t < time_ubound:
        i += 1

        # Determine conditions
        T_g = _temperature_gas(t)

        # Determine thermal properties
        c_a = c_a_T(T_g)
        rho_a = rho_a_T(T_g)
        lambda_pt = lambda_pt_T(T_g)
        conductivity_protection[i] = lambda_pt

        # Determine time step
        dt = _dt(c_a, rho_a, lambda_pt, d_p, A_p, V)
        dt = min([dt, time_step_lbound])
        time_rate.append(dt)

        # Determine current time
        time_.append(time_[i-1] + dt)
        t = time_[i]

        # Determine temperature change rate of the steel
        lambda_ave, theta_t, theta_at = lambda_pt, T_g, temperature_steel[i-1]
        temperature_rate_steel.append(_d_theta_at(c_a, rho_a, lambda_ave, d_p, A_p, V, theta_t, theta_at, dt))
        temperature_steel.append(temperature_steel[i-1] + temperature_rate_steel[i])

    # Convert lists to ndarray
    time_, time_rate = np.asarray(time_), np.asarray(time_rate)
    temperature_steel, temperature_rate_steel = np.asarray(temperature_steel), np.asarray(temperature_rate_steel)
    data_trivial = {""}

    return time_, temperature_steel, time_rate, temperature_rate_steel



import numpy as np


def solver_exposed(
        dx: np.float64,
        U_solid: np.float64,
        U_gas: np.float64,
        rho: np.float64,
        c: np.float64,
        h: np.float64,
        e: np.float64,
        dt: np.float64
):
    u0 = U_gas - U_solid
    u0_4 = U_gas ** 4 - U_solid ** 4
    e0 = u0 * h
    e2 = u0_4 * e
    e3 = rho * c * dx
    dU_dt_0 = (e0 + e2) / e3
    U_solid = U_solid + dU_dt_0 * dt

    return U_solid


def solver_solid(
        dx: np.ndarray,
        U: np.ndarray,
        k: np.ndarray,
        rho: np.ndarray,
        c: np.ndarray,
        dt: float
):
    """
    Description
    1-d heat transfer, finite difference, non-propagation approach, entire U at next time step n+1 is based upon U at
    the current temperature U.

    Dimension Indexing
    m: space dimension in x-axis, x[m=0] origin, x[m=X] end.
    n: time dimension, t[n=0] is beginning, t[n=T] is end.

    :param dx: length dimension in x-axis direction of each nodes
    :param U: temperature at each nodes
    :param k: thermal conductivities at each nodes
    :param rho: density at each nodes
    :param c: thermal heat capacity at each nodes
    :param dt: change in time
    :return U: new temperature at each nodes
    """

    U_new = np.zeros_like(U)

    du1 = U[0] - U[1]
    e1 = - du1 * k[0]
    en = rho[0] * c[0] * dx[0]
    U_new[0] = U[0] + e1 / en * dt

    du0 = U[-2] - U[-1]
    e1 = du0 * k[-1]
    en = rho[-1] * c[-1] * dx[-1]
    U_new[-1] = U[-1] + e1 / en * dt

    for m in range(1, len(dx) - 1):
        du0 = U[m - 1] - U[m]
        du1 = U[m] - U[m + 1]
        e1 = du0 * k[m] - du1 * k[m]  # todo: thermal properties can be averaged between adjacent nodes, not on current
        en = rho[m] * c[m] * dx[m]
        du_dt = e1 / en
        U_new[m] = U[m] + du_dt * dt

    return U_new



# -*- coding: utf-8 -*-
import os
import pandas as pd


def concatenate_dataframe_header(header_name:str, **dict_df:pd.DataFrame):

    dict_res = dict()

    for k, df in dict_df.items():
        dict_res[k] = df[header_name].values

    return pd.DataFrame.from_dict(dict_res)


def concatenate_dataframe_header2(*header_names:str, **dict_df:pd.DataFrame):

    dict_res = dict()

    for header_name in header_names:
        dict_res[header_name] = []

    for k, df in dict_df.items():
        for header_name in header_names:
            dict_res[header_name] += list(df[header_name].values)

    return pd.DataFrame.from_dict(dict_res)


def extract_results(*list_path_csv:str):
    # get out.csv path

    dict_df = dict()

    for path_csv in list_path_csv:
        k = os.path.basename(path_csv).replace('_out.csv', '')
        df = pd.DataFrame.from_csv(path_csv)
        dict_df[k] = df

    dict_df_out = dict()
    dict_df_out['solver_time_equivalence_solved'] = concatenate_dataframe_header('solver_time_equivalence_solved', **dict_df)
    dict_df_out['solver_time_equivalence_solved_merged'] = concatenate_dataframe_header2('solver_time_equivalence_solved', 'probability_weight', **dict_df)

    return dict_df_out



# -*- coding: utf-8 -*-
import copy
import warnings
from typing import Union

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from sfeprapy.func.fire_parametric_ec import fire as _fire_param
from sfeprapy.func.fire_parametric_ec_din import fire as _fire_param_ger
from sfeprapy.func.fire_travelling import fire as fire_travelling
from sfeprapy.func.heat_transfer_protected_steel_ec import protected_steel_eurocode as _steel_temperature
from sfeprapy.func.heat_transfer_protected_steel_ec import \
    protected_steel_eurocode_max_temperature as _steel_temperature_max


def _fire_travelling(**kwargs):
    if isinstance(kwargs['beam_location_length_m'], list) or isinstance(kwargs['beam_location_length_m'], np.ndarray):

        kwarg_ht_ec = dict(
            time=kwargs['t'],
            rho_steel=7850,
            area_steel_section=0.017,
            k_protection=0.2,
            rho_protection=800,
            c_protection=1700,
            thickness_protection=0.005,
            perimeter_protected=2.14,
        )

        temperature_steel_list = list()
        temperature_gas_list = fire_travelling(**kwargs)

        for temperature in temperature_gas_list:
            kwarg_ht_ec['temperature_ambient'] = temperature + 273.15
            temperature_steel_list.append(_steel_temperature_max(**kwarg_ht_ec))

        return temperature_gas_list[np.argmax(temperature_steel_list)] + 273.15, \
               kwargs['beam_location_length_m'][[np.argmax(temperature_steel_list)]][0]

    elif isinstance(kwargs['beam_location_length_m'], float) or isinstance(kwargs['beam_location_length_m'], int):

        return fire_travelling(**kwargs) + 273.15, kwargs['beam_location_length_m']


def decide_fire(
        window_height: float,
        window_width: float,
        window_open_fraction: float,
        room_breadth: float,
        room_depth: float,
        room_height: float,
        fire_mode: int,
        fire_load_density: float,
        fire_combustion_efficiency: float,
        fire_hrr_density: float,
        fire_spread_speed: float,
        **_
) -> dict:
    """Calculates equivalent time exposure for a protected steel element member in more realistic fire environment
    opposing to the standard fire curve ISO 834.

    PARAMETERS:
    :param window_height: [m], weighted window opening height
    :param window_width: [m], total window opening width
    :param window_open_fraction: [-], a factor is multiplied with the given total window opening area
    :param room_breadth: [m], room breadth (shorter direction of the floor plan)
    :param room_depth: [m], room depth (longer direction of the floor plan)
    :param room_height: [m], room height from floor to soffit (structural), disregard any non fire resisting floors
    :param fire_hrr_density: [MW m-2], fire maximum release rate per unit area
    :param fire_load_density:
    :param fire_combustion_efficiency:
    :param fire_spread_speed: [m s-1], TRAVELLING FIRE, fire spread speed
    :param fire_mode: 0 - parametric, 1 - travelling, 2 - ger parametric, 3 - (0 & 1), 4 (1 & 2)
    :return:
    EXAMPLE:
    """

    # PERMEABLE AND INPUT CHECKS

    fire_load_density_deducted = fire_load_density * fire_combustion_efficiency

    # Total window opening area
    window_area = window_height * window_width * window_open_fraction

    # Room floor area
    room_floor_area = room_breadth * room_depth

    # Room internal surface area, total, including window openings
    room_total_area = (2 * room_floor_area) + ((room_breadth + room_depth) * 2 * room_height)

    # Fire load density related to the total surface area A_t
    fire_load_density_total = fire_load_density_deducted * room_floor_area / room_total_area

    # Opening factor
    opening_factor = window_area * np.sqrt(window_height) / room_total_area

    # Spread speed - Does the fire spread to involve the full compartment?
    fire_spread_entire_room_time = room_depth / fire_spread_speed
    burn_out_time = max([fire_load_density_deducted / fire_hrr_density, 900.])

    if fire_mode == 0 or fire_mode == 1 or fire_mode == 2:  # enforced to ec parametric fire

        fire_type = fire_mode
    elif fire_mode == 3:  # enforced to ec parametric + travelling
        if fire_spread_entire_room_time < burn_out_time and 0.01 < opening_factor <= 0.2 and 50 <= fire_load_density_total <= 1000:
            fire_type = 0  # parametric fire
        else:  # Otherwise, it is a travelling fire
            fire_type = 1  # travelling fire
    elif fire_mode == 4:  # enforced to german parametric + travelling
        # If fire spreads throughout compartment and ventilation is within EC limits = Parametric fire
        if fire_spread_entire_room_time < burn_out_time and 0.125 <= (window_area / room_floor_area) <= 0.5:
            fire_type = 2  # german parametric
        else:  # Otherwise, it is a travelling fire
            fire_type = 1  # travelling fire
    else:
        raise ValueError('Unknown fire mode {fire_mode}.'.format(fire_mode=fire_mode))

    results = dict(
        fire_type=fire_type
    )

    return results


def evaluate_fire_temperature(
        window_height: float,
        window_width: float,
        window_open_fraction: float,
        room_breadth: float,
        room_depth: float,
        room_height: float,
        room_wall_thermal_inertia: float,
        fire_tlim: float,
        fire_type: float,
        fire_time: Union[list, np.ndarray],
        fire_nft_limit: float,
        fire_load_density: float,
        fire_combustion_efficiency: float,
        fire_hrr_density: float,
        fire_spread_speed: float,
        fire_t_alpha: float,
        fire_gamma_fi_q: float,
        beam_position_vertical: float,
        beam_position_horizontal: Union[np.ndarray, list, float] = -1.,
        **_
) -> dict:
    """Calculate temperature array of pre-defined fire type `fire_type`.

    PARAMETERS:
    :param window_height: [m], weighted window opening height
    :param window_width: [m], total window opening width
    :param window_open_fraction: [-], a factor is multiplied with the given total window opening area
    :param room_breadth: [m], room breadth (shorter direction of the floor plan)
    :param room_depth: [m], room depth (longer direction of the floor plan)
    :param room_height: [m], room height from floor to soffit (structural), disregard any non fire resisting floors
    :param room_wall_thermal_inertia: [J m-2 K-1 s-1/2], thermal inertia of room lining material
    :param fire_tlim: [s], PARAMETRIC FIRE, see parametric fire function for details
    :param fire_type: [-],
    :param fire_time: [K],
    :param fire_load_density:
    :param fire_combustion_efficiency:
    :param fire_t_alpha:
    :param fire_gamma_fi_q:
    :param beam_position_vertical:
    :param fire_hrr_density: [MW m-2], fire maximum release rate per unit area
    :param fire_spread_speed: [m s-1], TRAVELLING FIRE, fire spread speed
    :param beam_position_horizontal: [s], beam location, will be solved for the worst case if less than 0.
    :param fire_nft_limit: [K], TRAVELLING FIRE, maximum temperature of near field temperature
    :return:
    EXAMPLE:
    """

    fire_load_density_deducted = fire_load_density * fire_combustion_efficiency


    # Total window opening area
    window_area = window_height * window_width * window_open_fraction

    # Room floor area
    room_floor_area = room_breadth * room_depth

    # Room internal surface area, total, including window openings
    room_total_area = (2 * room_floor_area) + ((room_breadth + room_depth) * 2 * room_height)

    if fire_type == 0:
        kwargs_fire_0_paramec = dict(
            t=fire_time,
            A_t=room_total_area,
            A_f=room_floor_area,
            A_v=window_area,
            h_eq=window_height,
            q_fd=fire_load_density_deducted * 1e6,
            lambda_=room_wall_thermal_inertia ** 2,
            rho=1,
            c=1,
            t_lim=fire_tlim,
            temperature_initial=20 + 273.15,
        )
        fire_temperature = _fire_param(**kwargs_fire_0_paramec)

    elif fire_type == 1:
        if beam_position_horizontal < 0:
            beam_position_horizontal = np.linspace(0.5 * room_depth, room_depth, 7)[1:-1]

        kwargs_fire_1_travel = dict(
            t=fire_time,
            fire_load_density_MJm2=fire_load_density_deducted,
            fire_hrr_density_MWm2=fire_hrr_density,
            room_length_m=room_depth,
            room_width_m=room_breadth,
            fire_spread_rate_ms=fire_spread_speed,
            beam_location_height_m=beam_position_vertical,
            beam_location_length_m=beam_position_horizontal,
            fire_nft_limit_c=fire_nft_limit - 273.15,
            opening_width_m=window_width,
            opening_height_m=window_height,
            opening_fraction=window_open_fraction,
        )
        fire_temperature, beam_position_horizontal = _fire_travelling(**kwargs_fire_1_travel)

        if beam_position_horizontal <= 0:
            raise ValueError("Beam position less or equal to 0.")

    elif fire_type == 2:
        kwargs_fire_2_param_din = dict(
            t_array_s=fire_time,
            A_w_m2=window_area,
            h_w_m2=window_height,
            A_t_m2=room_total_area,
            A_f_m2=room_floor_area,
            t_alpha_s=fire_t_alpha,
            b_Jm2s05K=room_wall_thermal_inertia,
            q_x_d_MJm2=fire_load_density_deducted,
            gamma_fi_Q=fire_gamma_fi_q
        )
        fire_temperature = _fire_param_ger(**kwargs_fire_2_param_din)

    else:
        fire_temperature = None

    results = dict(
        fire_temperature=fire_temperature,
        beam_position_horizontal=beam_position_horizontal
    )

    return results


def solve_time_equivalence(
        fire_time: Union[list, np.ndarray],
        fire_temperature: Union[list, np.ndarray],
        beam_cross_section_area: float,
        beam_rho: float,
        protection_k: float,
        protection_rho: float,
        protection_c: float,
        protection_protected_perimeter: float,
        fire_time_iso834: Union[list, np.ndarray],
        fire_temperature_iso834: Union[list, np.ndarray],
        solver_temperature_goal: float,
        solver_max_iter: int,
        solver_thickness_ubound: float,
        solver_thickness_lbound: float,
        solver_tol: float,
        phi_teq: float,
        **_
) -> dict:
    """Calculates equivalent time exposure for a protected steel element member in more realistic fire environment
    opposing to the standard fire curve ISO 834.

    PARAMETERS:
    :param fire_time: [s], time array
    :param fire_temperature: [K], temperature array
    :param beam_cross_section_area: [m2], the steel beam element cross section area
    :param beam_rho: [kg/m3], steel beam element density
    :param solver_temperature_goal: [K], steel beam element expected failure temperature
    :param protection_k: steel beam element protection material thermal conductivity
    :param protection_rho: steel beam element protection material density
    :param protection_c: steel beam element protection material specific heat
    :param protection_protected_perimeter: [m], steel beam element protection material perimeter
    :param fire_time_iso834: [s], the time (array) component of ISO 834 fire curve
    :param fire_temperature_iso834: [K], the temperature (array) component of ISO 834 fire curve
    :param solver_max_iter: Maximum allowable iteration counts for seeking solution for time equivalence
    :param solver_thickness_ubound: [m], protection layer thickness upper bound initial condition for solving time equivalence
    :param solver_thickness_lbound: [m], protection layer thickness lower bound initial condition for solving time equivalence
    :param solver_tol: [K], tolerance for solving time equivalence
    :param phi_teq: [-], model uncertainty factor
    :return results: dict
    EXAMPLE:
    """

    # ============================================
    # GOAL SEEK TO MATCH STEEL FAILURE TEMPERATURE
    # ============================================

    # MATCH PEAK STEEL TEMPERATURE BY ADJUSTING PROTECTION LAYER THICKNESS

    solver_iter_count = 0  # count how many iterations for  the seeking process
    solver_convergence_status = False  # flag used to indicate when the seeking is successful

    # Default values
    solver_time_equivalence_solved = -1
    solver_steel_temperature_solved = -1
    solver_protection_thickness = -1

    if solver_temperature_goal > 0:  # check seeking temperature, opt out if less than 0

        # Solve heat transfer using EC3 correlations
        # SI UNITS FOR INPUTS!
        kwarg_ht_ec = dict(
            time=fire_time,
            temperature_ambient=fire_temperature,
            rho_steel=beam_rho,
            area_steel_section=beam_cross_section_area,
            k_protection=protection_k,
            rho_protection=protection_rho,
            c_protection=protection_c,
            perimeter_protected=protection_protected_perimeter,
            terminate_max_temperature=solver_temperature_goal + 5 * solver_tol,
        )

        # SOLVER START FROM HERE
        # ======================

        # Minimise f(x) - θ
        # where:
        #   f(x)    is the steel maximum temperature.
        #   x       is the thickness,
        #   θ       is the steel temperature goal

        time_at_max_temperature = fire_time[np.argmax(fire_temperature)]

        def f_(x, terminate_check_wait_time):
            kwarg_ht_ec["thickness_protection"] = x
            T_ = _steel_temperature_max(**kwarg_ht_ec, terminate_check_wait_time=terminate_check_wait_time)
            return T_

        # Check whether there is a solution within predefined protection thickness boundaries
        x1, x2 = solver_thickness_lbound, solver_thickness_ubound
        y1, y2 = f_(x1, time_at_max_temperature), f_(x2, time_at_max_temperature)
        t1, t2 = solver_temperature_goal - solver_tol, solver_temperature_goal + solver_tol

        if y2 <= solver_temperature_goal <= y1:

            while True:

                solver_iter_count += 1

                # Work out linear equation: f(x) = y = a x + b
                a = (y1 - y2) / (x1 - x2)
                b = y1 - a * x1

                # work out new y based upon interpolated y
                x3 = solver_protection_thickness = (solver_temperature_goal - b) / a
                y3 = solver_steel_temperature_solved = f_(x3, time_at_max_temperature)

                if x1 < 0 or x2 < 0 or x3 < 0:
                    print('check')

                if y3 < t1:  # steel temperature is too low, decrease thickness
                    x2 = x3
                    y2 = y3
                elif y3 > t2:  # steel temperature is too high, increase thickness
                    x1 = x3
                    y1 = y3
                else:
                    solver_convergence_status = True

                if solver_convergence_status:

                    # CALCULATE BEAM FIRE RESISTANCE PERIOD IN ISO 834
                    # ================================================
                    # Make steel time-temperature curve when exposed to the given ambient temperature, i.e. ISO 834.

                    kwarg_ht_ec["time"] = fire_time_iso834
                    kwarg_ht_ec["temperature_ambient"] = fire_temperature_iso834
                    kwarg_ht_ec["thickness_protection"] = x3
                    steel_temperature = _steel_temperature(**kwarg_ht_ec)

                    steel_time = np.concatenate((np.array([0]), fire_time_iso834, np.array([fire_time_iso834[-1]])))
                    steel_temperature = np.concatenate((np.array([-1]), steel_temperature, np.array([1e12])))
                    func_teq = interp1d(steel_temperature, steel_time, kind="linear", bounds_error=False, fill_value=-1)
                    solver_time_equivalence_solved = func_teq(solver_temperature_goal)

                    break

                elif solver_iter_count >= solver_max_iter:  # Terminate if maximum solving iteration is reached

                    if solver_temperature_goal > y3:
                        solver_time_equivalence_solved = -np.inf
                    elif solver_temperature_goal < y3:
                        solver_time_equivalence_solved = np.inf
                    else:
                        solver_time_equivalence_solved = np.nan  # theoretically impossible, for syntax error only
                    break

        elif solver_temperature_goal - 2 <= y1 <= solver_temperature_goal + 2:
            solver_protection_thickness = x1
            solver_steel_temperature_solved = y1
            solver_convergence_status = True

            # CALCULATE BEAM FIRE RESISTANCE PERIOD IN ISO 834
            # ================================================
            # Make steel time-temperature curve when exposed to the given ambient temperature, i.e. ISO 834.

            kwarg_ht_ec["time"] = fire_time_iso834
            kwarg_ht_ec["temperature_ambient"] = fire_temperature_iso834
            kwarg_ht_ec["thickness_protection"] = x1
            steel_temperature = _steel_temperature(**kwarg_ht_ec)

            steel_time = np.concatenate((np.array([0]), fire_time_iso834, np.array([fire_time_iso834[-1]])))
            steel_temperature = np.concatenate((np.array([-1]), steel_temperature, np.array([1e12])))
            func_teq = interp1d(steel_temperature, steel_time, kind="linear", bounds_error=False, fill_value=-1)
            solver_time_equivalence_solved = func_teq(solver_temperature_goal)

        elif solver_temperature_goal - 2 <= y2 <= solver_temperature_goal + 2:
            solver_protection_thickness = x2
            solver_steel_temperature_solved = y2
            solver_convergence_status = True

            # CALCULATE BEAM FIRE RESISTANCE PERIOD IN ISO 834
            # ================================================
            # Make steel time-temperature curve when exposed to the given ambient temperature, i.e. ISO 834.

            kwarg_ht_ec["time"] = fire_time_iso834
            kwarg_ht_ec["temperature_ambient"] = fire_temperature_iso834
            kwarg_ht_ec["thickness_protection"] = x2
            steel_temperature = _steel_temperature(**kwarg_ht_ec)

            steel_time = np.concatenate((np.array([0]), fire_time_iso834, np.array([fire_time_iso834[-1]])))
            steel_temperature = np.concatenate((np.array([-1]), steel_temperature, np.array([1e12])))
            func_teq = interp1d(steel_temperature, steel_time, kind="linear", bounds_error=False, fill_value=-1)
            solver_time_equivalence_solved = func_teq(solver_temperature_goal)

        # No solution, thickness upper bound is not thick enough
        elif solver_temperature_goal > y1:
            solver_protection_thickness = x1
            solver_steel_temperature_solved = y1
            solver_time_equivalence_solved = -np.inf

        # No solution, thickness lower bound is not thin enough
        elif solver_temperature_goal < y2:
            solver_protection_thickness = x2
            solver_steel_temperature_solved = y2
            solver_time_equivalence_solved = np.inf

    solver_time_equivalence_solved *= phi_teq

    results = dict(
        solver_convergence_status=solver_convergence_status,
        solver_time_equivalence_solved=solver_time_equivalence_solved,
        solver_steel_temperature_solved=solver_steel_temperature_solved,
        solver_protection_thickness=solver_protection_thickness,
        solver_iter_count=solver_iter_count,
    )

    return results


def mcs_out_post(df: pd.DataFrame) -> pd.DataFrame:
    df_res = copy.copy(df)
    df_res = df_res.replace(to_replace=[np.inf, -np.inf], value=np.nan)
    df_res = df_res.dropna(axis=0, how="any")

    dict_ = dict()
    dict_['fire_type'] = str({k: np.sum(df_res['fire_type'].values == k) for k in np.unique(df_res['fire_type'].values)})

    for k in ['beam_position_horizontal', 'fire_combustion_efficiency', 'fire_hrr_density', 'fire_load_density',
              'fire_nft_limit', 'fire_spread_speed', 'window_open_fraction', 'phi_teq', 'timber_fire_load']:
        try:
            x = df_res[k].values
            x1, x2, x3 = np.min(x), np.mean(x), np.max(x)
            dict_[k] = f'{x1:<9.3f} {x2:<9.3f} {x3:<9.3f}'
        except KeyError:
            pass

    list_ = [f'{k:<24.24}: {v}' for k, v in dict_.items()]

    print('\n'.join(list_), '\n')

    return df


def teq_main_wrapper(args):
    try:
        kwargs, q = args
        q.put("index: {}".format(kwargs["index"]))
        return teq_main(**kwargs)
    except (ValueError, AttributeError):
        return teq_main(**args)


def teq_main(
        case_name: str,
        n_simulations: int,
        probability_weight: float,
        index: int,
        beam_cross_section_area: float,
        beam_position_vertical: float,
        beam_position_horizontal: float,
        beam_rho: float,
        fire_time_duration: float,
        fire_time_step: float,
        fire_combustion_efficiency: float,
        fire_gamma_fi_q: float,
        fire_hrr_density: float,
        fire_load_density: float,
        fire_mode: int,
        fire_nft_limit: float,
        fire_spread_speed: float,
        fire_t_alpha: float,
        fire_tlim: float,
        protection_c: float,
        protection_k: float,
        protection_protected_perimeter: float,
        protection_rho: float,
        room_breadth: float,
        room_depth: float,
        room_height: float,
        room_wall_thermal_inertia: float,
        solver_temperature_goal: float,
        solver_max_iter: int,
        solver_thickness_lbound: float,
        solver_thickness_ubound: float,
        solver_tol: float,
        window_height: float,
        window_open_fraction: float,
        window_width: float,
        window_open_fraction_permanent: float,
        phi_teq: float = 1.,
        timber_charring_rate: float = None,
        timber_hc: float = None,
        timber_density: float = None,
        timber_exposed_area: float = None,
        timber_solver_tol: float = None,
        timber_solver_ilim: float = None,
        **_
) -> dict:
    # Make the longest dimension between (room_depth, room_breadth) as room_depth
    if room_depth < room_breadth:
        room_depth += room_breadth
        room_breadth = room_depth - room_breadth
        room_depth -= room_breadth

    window_open_fraction = window_open_fraction * (1 - window_open_fraction_permanent) + window_open_fraction_permanent

    # Calculate fire time, this is used for all fire curves in the calculation

    fire_time = np.arange(0, fire_time_duration + fire_time_step, fire_time_step)

    # Calculate ISO 834 fire temperature

    fire_time_iso834 = fire_time
    fire_temperature_iso834 = (345. * np.log10((fire_time / 60.) * 8. + 1.) + 20.) + 273.15  # in [K]

    # Inject results, i.e. these values will be in the output
    res = dict(
        case_name=case_name,
        n_simulations=n_simulations,
        probability_weight=probability_weight,
        index=index,
        # beam_cross_section_area=beam_cross_section_area,
        # beam_position_vertical=beam_position_vertical,
        beam_position_horizontal=beam_position_horizontal,
        # beam_rho=beam_rho,
        # fire_time=fire_time,
        fire_combustion_efficiency=fire_combustion_efficiency,
        # fire_gamma_fi_q=fire_gamma_fi_q,
        fire_hrr_density=fire_hrr_density,
        fire_load_density=fire_load_density,
        # fire_mode=fire_mode,
        fire_nft_limit=fire_nft_limit,
        fire_spread_speed=fire_spread_speed,
        # fire_t_alpha=fire_t_alpha,
        # fire_tlim=fire_tlim,
        # fire_temperature_iso834=fire_temperature_iso834,
        # fire_time_iso834=fire_time_iso834,
        # protection_c=protection_c,
        # protection_k=protection_k,
        # protection_protected_perimeter=protection_protected_perimeter,
        # protection_rho=protection_rho,
        # room_breadth=room_breadth,
        # room_depth=room_depth,
        # room_height=room_height,
        # room_wall_thermal_inertia=room_wall_thermal_inertia,
        # solver_temperature_goal=solver_temperature_goal,
        # solver_max_iter=solver_max_iter,
        # solver_thickness_lbound=solver_thickness_lbound,
        # solver_thickness_ubound=solver_thickness_ubound,
        # solver_tol=solver_tol,
        # window_height=window_height,
        window_open_fraction=window_open_fraction,
        # window_width=window_width,
        phi_teq=phi_teq,
        # timber_charring_rate=timber_charring_rate,
        # timber_hc=timber_hc,
        # timber_density=timber_density,
        # timber_exposed_area=timber_exposed_area,
        # timber_solver_tol=timber_solver_tol,
        # timber_solver_ilim=timber_solver_ilim,
    )

    # UNITS CONVERSATION
    timber_charring_rate *= 1 / 1000  # [mm/min] -> [m/min]
    timber_charring_rate *= 1 / 60  # [m/min] -> [m/s]

    # initial timber exposure time
    if timber_exposed_area > 0:
        timber_exposed_duration = 1200
    else:
        timber_exposed_duration = 0

    timber_solver_iter = -1  # initialise solver iteration count for timber fuel contribution

    while True:

        timber_solver_iter += 1

        timber_charred_depth = timber_charring_rate * timber_exposed_duration
        timber_charred_volume = timber_charred_depth * timber_exposed_area
        timber_charred_mass = timber_density * timber_charred_volume
        timber_fire_load = timber_charred_mass * timber_hc
        timber_fire_load_density = timber_fire_load / (room_breadth * room_depth)

        # To check what design fire to use

        res_decide_fire = decide_fire(
            window_height,
            window_width,
            window_open_fraction,
            room_breadth,
            room_depth,
            room_height,
            fire_mode,
            fire_load_density + timber_fire_load_density,
            fire_combustion_efficiency,
            fire_hrr_density,
            fire_spread_speed)

        # To calculate design fire temperature

        res_evaluate_fire_temperature = evaluate_fire_temperature(
            window_height,
            window_width,
            window_open_fraction,
            room_breadth,
            room_depth,
            room_height,
            room_wall_thermal_inertia,
            fire_tlim,
            res_decide_fire['fire_type'],
            fire_time,
            fire_nft_limit,
            fire_load_density + timber_fire_load_density,
            fire_combustion_efficiency,
            fire_hrr_density,
            fire_spread_speed,
            fire_t_alpha,
            fire_gamma_fi_q,
            beam_position_vertical,
            beam_position_horizontal)

        # To calculate time equivalence

        res_solve_time_equivalence = solve_time_equivalence(
            fire_time,
            res_evaluate_fire_temperature['fire_temperature'],
            beam_cross_section_area,
            beam_rho,
            protection_k,
            protection_rho,
            protection_c,
            protection_protected_perimeter,
            fire_time_iso834,
            fire_temperature_iso834,
            solver_temperature_goal,
            solver_max_iter,
            solver_thickness_ubound,
            solver_thickness_lbound,
            solver_tol,
            phi_teq)

        if timber_exposed_area <= 0 or timber_exposed_area is None:  # no timber exposed
            break
        elif not res_solve_time_equivalence['solver_convergence_status']:  # no time equivalence solution
            break
        elif timber_solver_iter >= timber_solver_ilim:  # over the solver iteration limit
            break
        elif abs(timber_exposed_duration - res_solve_time_equivalence[
            'solver_time_equivalence_solved']) <= timber_solver_tol:  # convergence sought successfully
            break
        else:
            timber_exposed_duration = res_solve_time_equivalence['solver_time_equivalence_solved']

    res_timber_solver = dict(
        timber_exposed_duration=timber_exposed_duration,
        timber_solver_iter=timber_solver_iter,
        timber_fire_load=timber_fire_load,
        timber_charred_mass=timber_charred_mass,
        timber_charred_volume=timber_charred_volume,
    )

    res.update(res_timber_solver)
    res.update(res_decide_fire)
    res.update(res_evaluate_fire_temperature)
    res.update(res_solve_time_equivalence)

    return res


def test_teq_phi():
    warnings.filterwarnings('ignore')

    from sfeprapy.func.fire_iso834 import fire as fire_iso834

    fire_time_ = np.arange(0, 2 * 60 * 60, 1)
    fire_temperature_iso834_ = fire_iso834(fire_time_, 293.15)

    input_param = dict(
        index=0,
        case_name='Standard 1',
        probability_weight=1,
        fire_time_step=30,
        fire_time_duration=5 * 60 * 60,
        n_simulations=1,
        beam_cross_section_area=0.017,
        beam_position_vertical=2.5,
        beam_position_horizontal=18,
        beam_rho=7850,
        fire_combustion_efficiency=0.8,
        fire_gamma_fi_q=1,
        fire_hrr_density=0.25,
        fire_load_density=420,
        fire_mode=0,
        fire_nft_limit=1050,
        fire_spread_speed=0.01,
        fire_t_alpha=300,
        fire_tlim=0.333,
        fire_temperature_iso834=fire_temperature_iso834_,
        fire_time_iso834=fire_time_,
        protection_c=1700,
        protection_k=0.2,
        protection_protected_perimeter=2.14,
        protection_rho=7850,
        room_breadth=16,
        room_depth=31.25,
        room_height=3,
        room_wall_thermal_inertia=720,
        solver_temperature_goal=620 + 273.15,
        solver_max_iter=20,
        solver_thickness_lbound=0.0001,
        solver_thickness_ubound=0.0500,
        solver_tol=1.,
        window_height=2,
        window_open_fraction=0.8,
        window_width=72,
        window_open_fraction_permanent=0,
        phi_teq=0.1,
        timber_charring_rate=0.7,
        timber_exposed_area=0,
        timber_hc=400,
        timber_density=500,
        timber_solver_ilim=20,
        timber_solver_tol=1,
    )

    input_param['phi_teq'] = 1.0
    teq_10 = teq_main(**input_param)['solver_time_equivalence_solved']

    input_param['phi_teq'] = 0.1
    teq_01 = teq_main(**input_param)['solver_time_equivalence_solved']

    assert abs(teq_10 / teq_01 - 10) < 0.001


def test():
    test_teq_phi()


if __name__ == '__main__':
    test()



#!/usr/bin/env python

import os
from codecs import open  # To use a consistent encoding

import setuptools

import sfeprapy

# Get the long description from the README file
with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'README.md')) as f:
    long_description = f.read()

setuptools.setup(

    name='sfeprapy',

    version=sfeprapy.__version__,

    description='Structural Fire Engineering - Probabilistic Reliability Assessment',

    author='Yan Fu',

    author_email='fuyans@gmail.com',

    url='https://github.com/fsepy/sfeprapy',

    download_url="https://github.com/fsepy/sfeprapy/archive/master.zip",

    keywords=["fire safety", "structural fire engineering"],

    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],

    long_description=long_description,

    long_description_content_type="text/markdown",

    packages=[
        'sfeprapy', 
        'sfeprapy.dat', 
        'sfeprapy.mc0', 
        'sfeprapy.mcs0',
        'sfeprapy.mcs1',
        'sfeprapy.mcs2',
        'sfeprapy.dist_fit', 
        'sfeprapy.func',
        'sfeprapy.func.heat_transfer_1d'
        ],

    install_requires=[
        'matplotlib>=2.2.2',
        'numpy>=1.15.0',
        'pandas>=0.23.3',
        'scipy>=1.1.0',
        'seaborn>=0.9.0',
        'tqdm',
        'xlrd',
    ],

    include_package_data=True,
)



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

from sfeprapy.dat.ec_3_1_2_kyT import ky2T_probabilistic_vectorised
from sfeprapy.func.fire_parametric_ec import fire as _fire_param
from sfeprapy.func.fire_parametric_ec_din import fire as _fire_param_ger
from sfeprapy.func.fire_travelling import fire as _fire_travelling
from sfeprapy.func.heat_transfer_protected_steel_ec import \
    protected_steel_eurocode_max_temperature as _steel_temperature_max


def a_solve_steel_protection_thickness_in_iso_fire(
        fire_iso834_time,
        fire_iso834_temperature,

        beam_rho,
        beam_cross_section_area,

        protection_k,
        protection_rho,
        protection_c,
        protection_protected_perimeter,

        solver_temperature_target,
        solver_fire_duration,
        solver_thickness_lbound,
        solver_thickness_ubound,
        solver_tolerance,
        solver_iteration_limit,

        **_
):
    # Solve heat transfer using EC3 correlations
    # SI UNITS FOR INPUTS!

    kwarg_ht_ec = dict(
        time=fire_iso834_time[fire_iso834_time <= solver_fire_duration],
        temperature_ambient=fire_iso834_temperature[fire_iso834_time <= solver_fire_duration],
        rho_steel=beam_rho,
        area_steel_section=beam_cross_section_area,
        k_protection=protection_k,
        rho_protection=protection_rho,
        c_protection=protection_c,
        perimeter_protected=protection_protected_perimeter,
        terminate_max_temperature=solver_temperature_target + 5 * solver_tolerance,
    )

    solver_iteration_count = 0  # count how many iterations for  the seeking process
    flag_solver_status = False  # flag used to indicate when the seeking is successful

    # Default values
    solver_thickness_solved = -1

    if solver_temperature_target > 0:  # check seeking temperature, opt out if less than 0

        # def f_(x):
        #     kwarg_ht_ec["thickness_protection"] = x  # NOTE! variable out function scope
        #     T_ = _steel_temperature_max(**kwarg_ht_ec, terminate_check_wait_time=iso834_time[np.argmax(iso834_temperature)])
        #     return T_

        # Check whether there is a solution within predefined protection thickness boundaries
        x1, x2 = solver_thickness_lbound, solver_thickness_ubound

        # y1, y2 = f_(x1), f_(x2)
        y1 = _steel_temperature_max(
            **kwarg_ht_ec,
            thickness_protection=x1,
            terminate_check_wait_time=solver_fire_duration - 30
        )
        y2 = _steel_temperature_max(
            **kwarg_ht_ec,
            thickness_protection=x2,
            terminate_check_wait_time=solver_fire_duration - 30
        )

        t1, t2 = solver_temperature_target - solver_tolerance, solver_temperature_target + solver_tolerance

        if (y2 - solver_tolerance) <= solver_temperature_target <= (y1 + solver_tolerance):

            while True:

                solver_iteration_count += 1

                # Work out linear equation: f(x) = y = a x + b
                a = (y1 - y2) / (x1 - x2)
                b = y1 - a * x1

                # work out new y based upon interpolated y
                x3 = solver_thickness_solved = (solver_temperature_target - b) / a
                # y3 = f_(x3)
                y3 = _steel_temperature_max(
                    **kwarg_ht_ec,
                    thickness_protection=x3,
                    terminate_check_wait_time=solver_fire_duration - 30
                )

                if y3 < t1:  # steel temperature is too low, decrease thickness
                    x2 = x3
                    y2 = y3
                elif y3 > t2:  # steel temperature is too high, increase thickness
                    x1 = x3
                    y1 = y3
                else:
                    flag_solver_status = True

                if flag_solver_status or (solver_iteration_count >= solver_iteration_limit):
                    break

        # No solution, thickness upper bound is not thick enough
        elif solver_temperature_target > y1:
            solver_thickness_solved = x1

        # No solution, thickness lower bound is not thin enough
        elif solver_temperature_target < y2:
            solver_thickness_solved = x2

    return solver_thickness_solved, flag_solver_status, solver_iteration_count


def b_calc_steel_temperature_in_design_fire(
        fire_time,
        fire_mode,

        room_depth,
        room_breadth,
        room_height,

        room_wall_thermal_inertia,

        window_height,
        window_width,
        window_open_fraction,

        beam_loc_z,
        beam_position,
        beam_rho,
        beam_cross_section_area,

        protection_k,
        protection_rho,
        protection_c,
        protection_protected_perimeter,
        protection_thickness,

        fire_load_density,
        fire_hrr_density,
        fire_tlim,
        fire_spread_speed,
        fire_nft_ubound,
        fire_t_alpha,
        fire_gamma_fi_q,

        **_
):
    # Make the longest dimension between (room_depth, room_breadth) as room_depth
    room_depth, room_breadth = max(room_depth, room_breadth), min(room_depth, room_breadth)

    # Total window opening area
    window_area = window_height * window_width * window_open_fraction

    # Room floor area
    room_floor_area = room_breadth * room_depth

    # Room internal surface area, total, including window openings
    room_total_area = (2 * room_floor_area) + ((room_breadth + room_depth) * 2 * room_height)

    # Fire load density related to the total surface area A_t
    fire_load_density_total = fire_load_density * room_floor_area / room_total_area

    # Opening factor
    opening_factor = window_area * np.sqrt(window_height) / room_total_area

    # Spread speed - Does the fire spread to involve the full compartment?
    fire_spread_entire_room_time = room_depth / fire_spread_speed
    burn_out_time = max([fire_load_density / fire_hrr_density, 900.])

    kwargs_fire_0_paramec = dict(
        t=fire_time,
        A_t=room_total_area,
        A_f=room_floor_area,
        A_v=window_area,
        h_eq=window_height,
        q_fd=fire_load_density * 1e6,
        lambda_=room_wall_thermal_inertia ** 2,
        rho=1,
        c=1,
        t_lim=fire_tlim,
        temperature_initial=20 + 273.15,
    )

    kwargs_fire_1_travel = dict(
        t=fire_time,
        fire_load_density_MJm2=fire_load_density,
        fire_hrr_density_MWm2=fire_hrr_density,
        room_length_m=room_depth,
        room_width_m=room_breadth,
        fire_spread_rate_ms=fire_spread_speed,
        beam_location_height_m=beam_loc_z,
        beam_location_length_m=beam_position,
        fire_nft_c=fire_nft_ubound,
        opening_width_m=window_width,
        opening_height_m=window_height,
        opening_fraction=window_open_fraction,
    )

    kwargs_fire_2_paramdin = dict(
        t_array_s=fire_time,
        A_w_m2=window_area,
        h_w_m2=window_height,
        A_t_m2=room_total_area,
        A_f_m2=room_floor_area,
        t_alpha_s=fire_t_alpha,
        b_Jm2s05K=room_wall_thermal_inertia,
        q_x_d_MJm2=fire_load_density,
        gamma_fi_Q=fire_gamma_fi_q
    )

    if fire_mode == 0:  # enforced to ec parametric fire

        # opening_factor = min(0.2, max(0.02, opening_factor))  # force opening factor fall in the boundary

        fire_temp = _fire_param(**kwargs_fire_0_paramec)
        fire_type = 0  # parametric fire

    elif fire_mode == 1:  # enforced to travelling fire

        fire_temp = _fire_travelling(**kwargs_fire_1_travel) + 273.15
        fire_type = 1  # travelling fire

    elif fire_mode == 2:  # enforced to german parametric fire

        fire_temp = _fire_param_ger(**kwargs_fire_2_paramdin)
        fire_type = 2  # german parametric

    elif fire_mode == 3:  # enforced to ec parametric + travelling
        if fire_spread_entire_room_time < burn_out_time and 0.02 < opening_factor <= 0.2 and 50 <= fire_load_density_total <= 1000:  # If fire spreads throughout compartment and ventilation is within EC limits = Parametric fire

            fire_temp = _fire_param(**kwargs_fire_0_paramec)
            fire_type = 0  # parametric fire

        else:  # Otherwise, it is a travelling fire

            fire_temp = _fire_travelling(**kwargs_fire_1_travel) + 273.15
            fire_type = 1  # travelling fire

    elif fire_mode == 4:  # enforced to german parametric + travelling

        if fire_spread_entire_room_time < burn_out_time and 0.125 <= (
                window_area / room_floor_area) <= 0.5:  # If fire spreads throughout compartment and ventilation is within EC limits = Parametric fire

            fire_temp = _fire_param_ger(**kwargs_fire_2_paramdin)
            fire_type = 2  # german parametric

        else:  # Otherwise, it is a travelling fire

            fire_temp = _fire_travelling(**kwargs_fire_1_travel) + 273.15
            fire_type = 1  # travelling fire

    else:
        raise ValueError('unknown fire_mode.')

    # Calculate maximum steel temperature
    # -----------------------------------

    kwarg_ht_ec = dict(
        time=fire_time,
        temperature_ambient=fire_temp,
        rho_steel=beam_rho,
        area_steel_section=beam_cross_section_area,
        k_protection=protection_k,
        rho_protection=protection_rho,
        c_protection=protection_c,
        thickness_protection=protection_thickness,
        perimeter_protected=protection_protected_perimeter,
    )

    solver_steel_temperature_solved = _steel_temperature_max(
        **kwarg_ht_ec,
        terminate_check_wait_time=fire_time[np.argmax(fire_temp)]
    )

    return solver_steel_temperature_solved, fire_type


def c_strength_reduction_factor(solver_steel_temperature_solved: np.ndarray, is_random_q:bool=True):

    if is_random_q:
        epsilon_q = np.random.random_sample(len(solver_steel_temperature_solved))
    else:
        epsilon_q = np.full(len(solver_steel_temperature_solved), 0.5)

    steel_strength_reduction_factor = ky2T_probabilistic_vectorised(
        T = solver_steel_temperature_solved,
        epsilon_q = epsilon_q
    )

    return steel_strength_reduction_factor


def main(
        df_mc_params: pd.DataFrame
):

    dict_mc_params = df_mc_params.to_dict(orient='index')

    # ==================================================================================================================
    # solve for protection thickness
    # ==================================================================================================================

    solver_thickness_solved, flag_solver_status, solver_iteration_count = \
        a_solve_steel_protection_thickness_in_iso_fire(**df_mc_params.loc[0].to_dict())

    _ = [df_mc_params.loc[0].to_dict()['solver_fire_duration'], solver_thickness_solved, flag_solver_status, solver_iteration_count]
    print('fire duration {:.0f}\nsolved thickness {:.10f}\nsolver status: {}\nsolver iter: {}\n'.format(*_))

    for uid, kwargs in dict_mc_params.items():

        solver_steel_temperature_solved, fire_type = \
            b_calc_steel_temperature_in_design_fire(protection_thickness=solver_thickness_solved, **kwargs)

        dict_mc_params[uid]['solver_thickness_solved'] = solver_thickness_solved
        dict_mc_params[uid]['flag_solver_status'] = flag_solver_status
        dict_mc_params[uid]['solver_iteration_count'] = solver_iteration_count
        dict_mc_params[uid]['fire_type'] = fire_type
        dict_mc_params[uid]['solver_steel_temperature_solved'] = solver_steel_temperature_solved

    df_out = pd.DataFrame.from_dict(dict_mc_params, orient='index')
    df_out.set_index("index", inplace=True)  # assign 'index' column as DataFrame index

    list_c_strength_reduction_factor = c_strength_reduction_factor(df_out['solver_steel_temperature_solved'].values, True)
    df_out['strength_reduction_factor'] = list_c_strength_reduction_factor

    # TODO: HASH ITEMS BELOW TO REMOVE FROM OUTPUT CSV
    df_out.pop('fire_time')
    df_out.pop('fire_iso834_time')
    df_out.pop('fire_iso834_temperature')
    df_out.pop("window_height")
    df_out.pop("window_width")
    # df_mcs_in.pop("room_breadth")
    # df_mcs_in.pop("room_depth")
    # df_mcs_in.pop("room_height")
    df_out.pop("room_wall_thermal_inertia")
    df_out.pop("fire_mode")
    df_out.pop("fire_time_step")
    df_out.pop("fire_tlim")
    df_out.pop("fire_hrr_density")
    df_out.pop("fire_duration")
    df_out.pop("fire_t_alpha")
    df_out.pop("fire_gamma_fi_q")
    df_out.pop("beam_rho")
    df_out.pop("beam_cross_section_area")
    df_out.pop("beam_loc_z")
    df_out.pop("protection_k")
    df_out.pop("protection_rho")
    df_out.pop("protection_c")
    df_out.pop("protection_protected_perimeter")
    df_out.pop("solver_temperature_target")
    df_out.pop("solver_fire_duration")
    df_out.pop("solver_thickness_lbound")
    df_out.pop("solver_thickness_ubound")
    df_out.pop("solver_tolerance")
    df_out.pop("solver_iteration_limit")
    # df_mcs_in.pop("solver_thickness_solved")
    # df_mcs_in.pop("flag_solver_status")
    # df_mcs_in.pop("solver_iteration_count")
    # df_mcs_in.pop("fire_type")

    return df_out

# if self.n_threads <= 1:
#     results = []
#     for dict_mc_params in mc_param_list:
#         results.append(calc_main(**dict_mc_params))
#
# else:
#     time_simulation_start = time.perf_counter()
#     m = mp.Manager()
#     q = m.Queue()
#     p = mp.Pool(n_threads, maxtasksperchild=1000)
#     jobs = p.map_async(calc_time_equivalence_worker, [(dict_mc_param, q) for dict_mc_param in mc_param_list])
#     n_steps = 24  # length of the progress bar
#     while True:
#         if jobs.ready():
#             time_simulation_consumed = time.perf_counter() - time_simulation_start
#             print("{}{} {:.1f}s ".format('█' * round(n_steps), '-' * round(0), time_simulation_consumed))
#             break
#         else:
#             path_ = q.qsize() / n_simulations * n_steps
#             print("{}{} {:03.1f}%".format('█' * int(round(path_)), '-' * int(n_steps - round(path_)),
#                                           path_ / n_steps * 100),
#                   end='\r')
#             time.sleep(1)
#     p.close()
#     p.join()
#     results = jobs.get()



# Equation 37: Mass flow rate of smoke from a vertical opening
def m_dot_smoke_func(
        C_e,
        p,
        h,
        w,
        C_d,
        rho_0,
        T_max,
        g,
        theta_max,
        T_0,
        kappa_m
):
    """
    :param C_e:
    :param p:
    :param h: Compartment opening.
    :param w:
    :param C_d: is a discharge coefficient for the opening, 0.6 for an opening with a deep down-stand or 1.0 for no down-stand.
    :param rho_0:
    :param T_max:
    :param g:
    :param theta_max:
    :param T_0:
    :param kappa_m: is a 'profile correction factor' to allow for departure from a step function for the vertical temperature gradient (kappa_m is 1.0 for a step function but can be taken as 1.3 for most hot gas layers where theta_max is less than 300 deg.C).
    :return:
    """
    aa = 2 / 3 * rho_0 / T_max
    bb = (2 * g * theta_max * T_0 * kappa_m) ** 0.5
    k = aa * bb

    a = C_e * p * h ** (1.5) * w
    b = w ** (2 / 3)
    c = 1 / (C_d)
    d = ((C_e * p) / k) ** (2 / 3)
    m_dot_smoke = a / (b + c * d) ** (1.5)

    return m_dot_smoke


# Equation 54. Theoretically-based design formula (utilizing an empirical entrainment coefficient from the same experiments) for the balcony spill plume with channelling screens.
def m_dot_smoke_func_54(
        eta, E, Q_dot_c, W_s, z_s, d_s
):
    # Equation 57. The virtual source.
    z_0 = 3 * d_s

    m_dot_smoke = eta * (2 * E) ** (2/3) * (Q_dot_c) ** (1/3) * (W_s) ** (2/3) * (z_s + z_0) * (1 + 2 * E * (z_s / W_s)) ** (2/3)
    return m_dot_smoke


# Equation 55. It is recommended in Air entrainment into balcony spill plumes [29], for the two-dimensional channelled balcony spill plume that E = 0.11 and eta = 0.42. Therefore, in the linear entrainment region (0 << z_s << 5 * W_s) for the two dimensional channelled balcony spill plume, equation 54 approximates to
def m_dot_smoke_func_55(
        Q_dot_c, W_s, z_s, d_s
):
    # Equation 57. The virtual source.
    z_0 = 3 * d_s

    m_dot_smoke = 0.15 * (Q_dot_c) ** (1/3) * (W_s) ** (2/3) * (z_s + z_0)
    return m_dot_smoke


# Eq uation 56. It is recommended in Air entrainment into balcony spill plumes [29], for the three dimensional channelled balcony spill plume that E = 0.13 and eta = 0.42. In the linear entrainment region (0 << z_s << 5 * W_s) for the three-dimensional channelled balcony spill plume, equation 54 approximates to.
def m_dot_smoke_func_56(
        Q_dot_c, W_s, z_s, d_s
):
    # Equation 57. The virtual source.
    z_0 = 3 * d_s

    m_dot_smoke = 0.17 * (Q_dot_c) ** (1/3) * (W_s) ** (2/3) * (z_s + z_0)
    return m_dot_smoke



# -*- coding: utf-8 -*-
import copy
import warnings
from typing import Union

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from sfeprapy.func.fire_parametric_ec import fire as _fire_param
from sfeprapy.func.fire_parametric_ec_din import fire as _fire_param_ger
from sfeprapy.func.fire_travelling import fire as fire_travelling
from sfeprapy.func.heat_transfer_protected_steel_ec import protected_steel_eurocode as _steel_temperature
from sfeprapy.func.heat_transfer_protected_steel_ec import \
    protected_steel_eurocode_max_temperature as _steel_temperature_max

warnings.filterwarnings('ignore')


def _fire_travelling(**kwargs):
    if isinstance(kwargs['beam_location_length_m'], list) or isinstance(kwargs['beam_location_length_m'], np.ndarray):

        kwarg_ht_ec = dict(
            time=kwargs['t'],
            temperature_ambient=None,
            rho_steel=7850,
            area_steel_section=0.017,
            k_protection=0.2,
            rho_protection=800,
            c_protection=1700,
            thickness_protection=0.005,
            perimeter_protected=2.14,
        )

        temperature_steel_list = list()
        temperature_gas_list = fire_travelling(**kwargs)

        for temperature in temperature_gas_list:
            kwarg_ht_ec['temperature_ambient'] = temperature + 273.15
            temperature_steel_list.append(_steel_temperature_max(**kwarg_ht_ec))

        return temperature_gas_list[np.argmax(temperature_steel_list)]+273.15, \
               kwargs['beam_location_length_m'][[np.argmax(temperature_steel_list)]][0]

    elif isinstance(kwargs['beam_location_length_m'], float):

        return fire_travelling(**kwargs)+273.15, \
               kwargs['beam_location_length_m']


def calc_time_equivalence_worker(arg):
    kwargs, q = arg
    result = calc_time_equivalence(**kwargs)
    q.put("index: {}".format(kwargs["index"]))
    return result


def calc_time_equivalence(
        case_name,
        time_step,
        time_limiting,
        window_height,
        window_width,
        window_open_fraction,
        room_breadth,
        room_depth,
        room_height,
        room_wall_thermal_inertia,
        fire_mode,
        fire_nft_ubound,
        fire_load_density,
        fire_combustion_efficiency,
        fire_hrr_density,
        fire_spread_speed,
        fire_duration,
        fire_t_alpha,
        fire_gamma_fi_q,
        beam_cross_section_area,
        beam_temperature_goal,
        beam_loc_z,
        protection_k,
        protection_rho,
        protection_c,
        protection_thickness,
        protection_protected_perimeter,
        iso834_time,
        iso834_temperature,
        probability_weight:int,
        index,
        beam_position: Union[np.ndarray, list, float] = -1,
        solver_max_iter=20,
        solver_thickness_ubound=0.0500,
        solver_thickness_lbound=0.0001,
        solver_tol=1.,
        **_
):
    """
    NAME: grouped_a_b
    AUTHOR: Ian Fu
    DATE: 11 March 2019
    DESCRIPTION:
    Calculates equivalent time exposure for a protected steel element member in more realistic fire environment
    opposing to the standard fire curve ISO 834.

    PARAMETERS:
    :param time_step: [s], time step used for numerical calculation
    :param time_start: [s], simulation starting time
    :param time_limiting: [s], PARAMETRIC FIRE, see parametric fire function for details
    :param window_height: [m], weighted window opening height
    :param window_width: [m], total window opening width
    :param window_open_fraction: [-], a factor is multiplied with the given total window opening area
    :param room_breadth: [m], room breadth (shorter direction of the floor plan)
    :param room_depth: [m], room depth (longer direction of the floor plan)
    :param room_height: [m], room height from floor to soffit (structural), disregard any non fire resisting floors
    :param room_wall_thermal_inertia: [J m-2 K-1 s-1/2], thermal inertia of room lining material
    :param fire_load_density_deducted: [MJ m-2], fire load per unit area
    :param fire_hrr_density: [MW m-2], fire maximum release rate per unit area
    :param fire_spread_speed: [m s-1], TRAVELLING FIRE, fire spread speed
    :param fire_duration: [s], simulation time
    :param beam_position: [s], beam location, will be solved for the worst case if less than 0.
    :param beam_rho: [kg m-3], density of the steel beam element
    :param beam_c: [?], specific heat of the steel element
    :param beam_cross_section_area: [m2], the steel beam element cross section area
    :param beam_temperature_goal: [K], steel beam element expected failure temperature
    :param protection_k: steel beam element protection material thermal conductivity
    :param protection_rho: steel beam element protection material density
    :param protection_c: steel beam element protection material specific heat
    :param protection_thickness: steel beam element protection material thickness
    :param protection_protected_perimeter: [m], steel beam element protection material perimeter
    :param iso834_time: [s], the time (array) component of ISO 834 fire curve
    :param iso834_temperature: [K], the temperature (array) component of ISO 834 fire curve
    :param fire_nft_ubound: [K], TRAVELLING FIRE, maximum temperature of near field temperature
    :param solver_max_iter: Maximum allowable iteration counts for seeking solution for time equivalence
    :param solver_thickness_ubound: [m], protection layer thickness upper bound initial condition for solving time equivalence
    :param solver_thickness_lbound: [m], protection layer thickness lower bound initial condition for solving time equivalence
    :param solver_tol: [K], tolerance for solving time equivalence
    :param index: will be returned for indicating the index of the current iteration (was used for multiprocessing)
    :param fire_mode: 0 - parametric, 1 - travelling, 2 - ger parametric, 3 - (0 & 1), 4 (1 & 2)
    :return:
    EXAMPLE:
    """

    # DO NOT CHANGE, LEGACY PARAMETERS
    # Used to define fire curve start time, depreciated on 11/03/2019 after introducing the DIN annex ec parametric
    # fire.
    time_start = 0
    beam_rho = 7850

    # PERMEABLE AND INPUT CHECKS

    fire_load_density_deducted = fire_load_density * fire_combustion_efficiency

    # Make the longest dimension between (room_depth, room_breadth) as room_depth
    room_depth, room_breadth = max(room_depth, room_breadth), min(room_depth, room_breadth)

    # Total window opening area
    window_area = window_height * window_width * window_open_fraction

    # Room floor area
    room_floor_area = room_breadth * room_depth

    # Room internal surface area, total, including window openings
    room_total_area = (2 * room_floor_area) + ((room_breadth + room_depth) * 2 * room_height)

    # Fire load density related to the total surface area A_t
    fire_load_density_total = fire_load_density_deducted * room_floor_area / room_total_area

    # Opening factor
    opening_factor = window_area * np.sqrt(window_height) / room_total_area

    # Spread speed - Does the fire spread to involve the full compartment?
    fire_spread_entire_room_time = room_depth / fire_spread_speed
    burn_out_time = max([fire_load_density_deducted / fire_hrr_density, 900.])

    # if beam_position is not provided, solve for the worst case
    if beam_position < 0:
        beam_position = np.linspace(0.5*room_depth, room_depth, 7)[1:-1]

    fire_time = np.arange(time_start, fire_duration + time_step, time_step, dtype=float)

    kwargs_fire_0_paramec = dict(
        t=fire_time,
        A_t=room_total_area,
        A_f=room_floor_area,
        A_v=window_area,
        h_eq=window_height,
        q_fd=fire_load_density_deducted * 1e6,
        lambda_=room_wall_thermal_inertia ** 2,
        rho=1,
        c=1,
        t_lim=time_limiting,
        temperature_initial=20 + 273.15,
    )

    kwargs_fire_1_travel = dict(
        t=fire_time,
        fire_load_density_MJm2=fire_load_density_deducted,
        fire_hrr_density_MWm2=fire_hrr_density,
        room_length_m=room_depth,
        room_width_m=room_breadth,
        fire_spread_rate_ms=fire_spread_speed,
        beam_location_height_m=beam_loc_z,
        beam_location_length_m=beam_position,
        fire_nft_limit_c=fire_nft_ubound,
        opening_width_m=window_width,
        opening_height_m=window_height,
        opening_fraction=window_open_fraction,
    )

    kwargs_fire_2_paramdin = dict(
        t_array_s=fire_time,
        A_w_m2=window_area,
        h_w_m2=window_height,
        A_t_m2=room_total_area,
        A_f_m2=room_floor_area,
        t_alpha_s=fire_t_alpha,
        b_Jm2s05K=room_wall_thermal_inertia,
        q_x_d_MJm2=fire_load_density_deducted,
        gamma_fi_Q=fire_gamma_fi_q
    )
    if fire_mode == 0:  # enforced to ec parametric fire

        # opening_factor = min(0.2, max(0.02, opening_factor))  # force opening factor fall in the boundary
    
        fire_temperature = _fire_param(**kwargs_fire_0_paramec)
        fire_type = 0  # parametric fire

    elif fire_mode == 1:  # enforced to travelling fire

        fire_temperature, beam_position = _fire_travelling(**kwargs_fire_1_travel)
        fire_type = 1  # travelling fire

    elif fire_mode == 2:  # enforced to german parametric fire

        fire_temperature = _fire_param_ger(**kwargs_fire_2_paramdin)
        fire_type = 2   # german parametric

    elif fire_mode == 3:  # enforced to ec parametric + travelling

        # print(fire_spread_entire_room_time < burn_out_time, 0.02 < opening_factor <= 0.2, 50 <= fire_load_density_total <= 1000)

        if fire_spread_entire_room_time < burn_out_time and 0.02 < opening_factor <= 0.2 and 50 <= fire_load_density_total <= 1000:  # If fire spreads throughout compartment and ventilation is within EC limits = Parametric fire

            fire_temperature = _fire_param(**kwargs_fire_0_paramec)
            fire_type = 0  # parametric fire

        else:  # Otherwise, it is a travelling fire

            fire_temperature, beam_position = _fire_travelling(**kwargs_fire_1_travel)
            fire_type = 1  # travelling fire

    elif fire_mode == 4:  # enforced to german parametric + travelling

        if fire_spread_entire_room_time < burn_out_time and 0.125 <= (window_area / room_floor_area) <= 0.5:  # If fire spreads throughout compartment and ventilation is within EC limits = Parametric fire

            fire_temperature = _fire_param_ger(**kwargs_fire_2_paramdin)
            fire_type = 2  # german parametric

        else:  # Otherwise, it is a travelling fire

            fire_temperature, beam_position = _fire_travelling(**kwargs_fire_1_travel)
            fire_type = 1  # travelling fire

    else:

        raise ValueError('Unknown fire mode {fire_mode}.'.format(fire_mode=fire_mode))

    if fire_temperature[0] == np.nan:

        warnings.warn('Abnormal fire temperature.')
        print(fire_temperature)

    if not isinstance(beam_position, float):
        beam_position = np.nan

    # ============================================
    # GOAL SEEK TO MATCH STEEL FAILURE TEMPERATURE
    # ============================================

    # HELPER FUNCTION: CALCULATE TEQ FOR A GIVEN PROTECTION THICKNESS

    def find_time_equivalence_for_a_thickness(thickness_protection_, kwarg_ht_ec_, iso834_time_, iso834_temperature_, func_find_steel_temperature_, temperature_goal_):

        # CALCULATE BEAM FIRE RESISTANCE PERIOD IN ISO 834
        # Make steel time-temperature curve when exposed to the given ambient temperature, i.e. ISO 834.

        kwarg_ht_ec_["time"] = iso834_time_
        kwarg_ht_ec_["temperature_ambient"] = iso834_temperature_
        kwarg_ht_ec_["thickness_protection"] = thickness_protection_
        if iso834_temperature_[0] == np.nan:
            print('sh')
        steel_temperature = func_find_steel_temperature_(**kwarg_ht_ec_)

        steel_time = np.concatenate((np.array([0]), iso834_time, np.array([iso834_time[-1]])))
        steel_temperature = np.concatenate((np.array([-1]), steel_temperature, np.array([1e12])))
        func_teq = interp1d(steel_temperature, steel_time, kind="linear", bounds_error=False, fill_value=-1)
        solver_time_equivalence_solved = func_teq(temperature_goal_)

        return solver_time_equivalence_solved

    # MATCH PEAK STEEL TEMPERATURE BY ADJUSTING PROTECTION LAYER THICKNESS

    solver_iter_count = 0  # count how many iterations for  the seeking process
    solver_convergence_status = False  # flag used to indicate when the seeking is successful

    # Default values
    solver_time_equivalence_solved = -1
    solver_steel_temperature_solved = -1
    solver_protection_thickness = -1

    if beam_temperature_goal > 0:  # check seeking temperature, opt out if less than 0

        # Solve heat transfer using EC3 correlations
        # SI UNITS FOR INPUTS!
        kwarg_ht_ec = dict(
            time=fire_time,
            temperature_ambient=fire_temperature,
            rho_steel=beam_rho,
            area_steel_section=beam_cross_section_area,
            k_protection=protection_k,
            rho_protection=protection_rho,
            c_protection=protection_c,
            perimeter_protected=protection_protected_perimeter,
            terminate_max_temperature=beam_temperature_goal + 5 * solver_tol,
        )

        # SOLVER START FROM HERE
        # ======================

        # Minimise f(x) - θ
        # where:
        #   f(x)    is the steel maximum temperature.
        #   x       is the thickness,
        #   θ       is the steel temperature goal

        time_at_max_temperature = fire_time[np.argmax(fire_temperature)]

        def f_(x, terminate_check_wait_time):
            kwarg_ht_ec["thickness_protection"] = x
            T_ = _steel_temperature_max(**kwarg_ht_ec, terminate_check_wait_time=terminate_check_wait_time)
            return T_

        # Check whether there is a solution within predefined protection thickness boundaries
        x1, x2 = solver_thickness_lbound, solver_thickness_ubound
        y1, y2 = f_(x1, time_at_max_temperature), f_(x2, time_at_max_temperature)
        t1, t2 = beam_temperature_goal - solver_tol, beam_temperature_goal + solver_tol

        # if (y2 - solver_tol) <= beam_temperature_goal <= (y1 + solver_tol):
        if y2 <= beam_temperature_goal <= y1:

            while True:

                solver_iter_count += 1

                # Work out linear equation: f(x) = y = a x + b
                a = (y1 - y2) / (x1 - x2)
                b = y1 - a * x1

                # work out new y based upon interpolated y
                x3 = solver_protection_thickness = (beam_temperature_goal - b) / a
                y3 = solver_steel_temperature_solved = f_(x3, time_at_max_temperature)

                if x1 < 0 or x2 < 0 or x3 < 0:
                    print('check')

                if y3 < t1:  # steel temperature is too low, decrease thickness
                    x2 = x3
                    y2 = y3
                elif y3 > t2:  # steel temperature is too high, increase thickness
                    x1 = x3
                    y1 = y3
                else:
                    solver_convergence_status = True

                if solver_convergence_status:

                    # CALCULATE BEAM FIRE RESISTANCE PERIOD IN ISO 834
                    # ================================================
                    # Make steel time-temperature curve when exposed to the given ambient temperature, i.e. ISO 834.

                    kwarg_ht_ec["time"] = iso834_time
                    kwarg_ht_ec["temperature_ambient"] = iso834_temperature
                    kwarg_ht_ec["thickness_protection"] = x3
                    steel_temperature = _steel_temperature(**kwarg_ht_ec)

                    steel_time = np.concatenate((np.array([0]), iso834_time, np.array([iso834_time[-1]])))
                    steel_temperature = np.concatenate((np.array([-1]), steel_temperature, np.array([1e12])))
                    func_teq = interp1d(steel_temperature, steel_time, kind="linear", bounds_error=False, fill_value=-1)
                    solver_time_equivalence_solved = func_teq(beam_temperature_goal)

                    break

                elif solver_iter_count >= solver_max_iter:  # Terminate if maximum solving iteration is reached

                    if beam_temperature_goal > y3:
                        solver_time_equivalence_solved = -np.inf
                    elif beam_temperature_goal < y3:
                        solver_time_equivalence_solved = np.inf
                    else:
                        solver_time_equivalence_solved = np.nan  # theoretically impossible, for syntax error only
                    break

        elif beam_temperature_goal - 2 <= y1 <= beam_temperature_goal + 2:
            solver_protection_thickness = x1
            solver_steel_temperature_solved = y1
            solver_convergence_status = True

            # CALCULATE BEAM FIRE RESISTANCE PERIOD IN ISO 834
            # ================================================
            # Make steel time-temperature curve when exposed to the given ambient temperature, i.e. ISO 834.

            kwarg_ht_ec["time"] = iso834_time
            kwarg_ht_ec["temperature_ambient"] = iso834_temperature
            kwarg_ht_ec["thickness_protection"] = x1
            steel_temperature = _steel_temperature(**kwarg_ht_ec)

            steel_time = np.concatenate((np.array([0]), iso834_time, np.array([iso834_time[-1]])))
            steel_temperature = np.concatenate((np.array([-1]), steel_temperature, np.array([1e12])))
            func_teq = interp1d(steel_temperature, steel_time, kind="linear", bounds_error=False, fill_value=-1)
            solver_time_equivalence_solved = func_teq(beam_temperature_goal)

        elif beam_temperature_goal - 2 <= y2 <= beam_temperature_goal + 2:
            solver_protection_thickness = x2
            solver_steel_temperature_solved = y2
            solver_convergence_status = True

            # CALCULATE BEAM FIRE RESISTANCE PERIOD IN ISO 834
            # ================================================
            # Make steel time-temperature curve when exposed to the given ambient temperature, i.e. ISO 834.

            kwarg_ht_ec["time"] = iso834_time
            kwarg_ht_ec["temperature_ambient"] = iso834_temperature
            kwarg_ht_ec["thickness_protection"] = x2
            steel_temperature = _steel_temperature(**kwarg_ht_ec)

            steel_time = np.concatenate((np.array([0]), iso834_time, np.array([iso834_time[-1]])))
            steel_temperature = np.concatenate((np.array([-1]), steel_temperature, np.array([1e12])))
            func_teq = interp1d(steel_temperature, steel_time, kind="linear", bounds_error=False, fill_value=-1)
            solver_time_equivalence_solved = func_teq(beam_temperature_goal)

        # No solution, thickness upper bound is not thick enough
        elif beam_temperature_goal > y1:
            solver_protection_thickness = x1
            solver_steel_temperature_solved = y1
            solver_time_equivalence_solved = -np.inf

        # No solution, thickness lower bound is not thin enough
        elif beam_temperature_goal < y2:
            solver_protection_thickness = x2
            solver_steel_temperature_solved = y2
            solver_time_equivalence_solved = np.inf

    results = dict(
        index=index,
        case_name=case_name,
        fire_type=fire_type,
        fire_load_density=fire_load_density,
        fire_combustion_efficiency=fire_combustion_efficiency,
        fire_hrr_density=fire_hrr_density,
        fire_spread_speed=fire_spread_speed,
        fire_nft_ubound=fire_nft_ubound,
        fire_time=fire_time,
        fire_temperature=fire_temperature,
        window_open_fraction=window_open_fraction,
        beam_position=beam_position,
        opening_factor=opening_factor,
        solver_convergence_status=solver_convergence_status,
        solver_time_equivalence_solved=solver_time_equivalence_solved,
        solver_steel_temperature_solved=solver_steel_temperature_solved,
        solver_protection_thickness=solver_protection_thickness,
        solver_iter_count=solver_iter_count,
        probability_weight=probability_weight,
    )

    return results


def y_results_summary(df_res: pd.DataFrame):

    df_res = copy.copy(df_res)
    df_res = df_res.replace(to_replace=[np.inf, -np.inf], value=np.nan)
    # df_res.replace(-np.inf, np.nan)
    df_res = df_res.dropna(axis=0, how="any")
    
    str_fmt = '{:<24.24}: {}\n'
    str_fmt2 = '{:<24.24}: {:<.3f}\n'

    str_out = ''

    fire_type = df_res['fire_type'].values
    fire_type_val = dict()
    for k in np.unique(fire_type):
        fire_type_val[k] = np.sum(fire_type == k)

    str_out += str_fmt.format('fire_type', fire_type_val)
    str_out += str_fmt.format('solver_convergence_status', np.sum(df_res['solver_convergence_status'].values))
    str_out += str_fmt2.format('beam_position', df_res['beam_position'].mean())
    str_out += str_fmt2.format('fire_combustion_efficiency', df_res['fire_combustion_efficiency'].mean())
    str_out += str_fmt2.format('fire_hrr_density', df_res['fire_hrr_density'].mean())
    str_out += str_fmt2.format('fire_load_density', df_res['fire_load_density'].mean())
    str_out += str_fmt2.format('fire_nft_ubound', df_res['fire_nft_ubound'].mean())
    str_out += str_fmt2.format('fire_spread_speed', df_res['fire_spread_speed'].mean())
    str_out += str_fmt2.format('window_open_fraction', df_res['window_open_fraction'].mean())

    return str_out
